
/*
   folder.c -- open mailbox for SMTP 
	
   Copyright (C) 1998, 1999  Xuming <xuming@iname.com>
   
   This program is free software; you can redistribute it and/or 
   modify it under the terms of the GNU General Public License as 
   published by the Free Software Foundation; either version 2, or 
   (at your option) any later version.
	   
   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of 
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License, the file COPYING in this directory, for
   more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/*
   This module handle the mail folder for sending mail. It defines
   two methods for open mail folder:
   
     open_folder(mbox)  open a specified mail folder to send.
     online_letter()  build a mail on the console or from the stream.
	 
   and three routines to access specified letter:
   
     active_mail(id)              specifies an active letter
     read_from_mail(buf, len)      read the text of the active letter
     mark_mail_remove(id)          mark this letter should be removed
     dump_tmp_folder()             dump temporary mail folder
     print_tmp_mails()             ditto, but with detail information
	 
   finally, do not forget to close the folder:
   
     close_folder()
*/
	 
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "defs.h"


static Folder  *rootfld;
static char    gdir[] = PMTMP;
static int     fid = 1;
static char    *signature = NULL;


static void folder_insert(Folder *tar)
{
    Folder *fo;
    
    if (rootfld == NULL)  {
	rootfld = tar;
	rootfld->next = NULL;
	rootfld->prev = NULL;
    } else {
	for (fo = rootfld; fo->next; fo = fo->next);
	fo->next = tar;
	tar->next = NULL;
	tar->prev = fo;
    }
}

static Folder *check_loaded(char *fname)
{
    Folder *fo;

    for (fo = rootfld; fo; fo = fo->next)  {
        if (!strcmp(fo->name, fname))  break;
    }
    return fo;
}


static void load_signature(char *s)
{
    FILE	*fsig;
    char	buf[SVRBUF], *sig;
    int		len;
    
    if (signature)  return;

    fullpath(s, buf);
    if ((fsig = fopen(buf, "r")) == NULL)  {
	logPrintf(LOG_WARN, 
		"can not open signature file. [%s]\n", buf);
	return;
    }
    
    fseek(fsig, 0, SEEK_END);
    if ((len = ftell(fsig)) > MAX_SIG)  {
	logPrintf(LOG_WARN, 
		"Too large the signature file is. [%d]\n", len);
	fclose(fsig);
	return;
    }
    
    /*fprintf(stderr, "%s [%d]\n", buf, len);*/
    
    rewind(fsig);
    if ((signature = (char*) malloc(len + len / 5)) == NULL)  {
	fclose(fsig);
	logPrintf(LOG_WARN, 
		"Insufficent memory for signature built. [%d]\n", len);
	return;
    }

    *buf = '>';
    sig = buf + 1;
    *signature = '\0';
    
    while (fgets(sig, SVRBUF - 1, fsig))  {
	if (!strncmp(sig, "From ", 5))  strcat(signature, buf);
	else  strcat(signature, sig);
    }
    fclose(fsig);
}	       
    

int folder_number()
{
    Folder *fo;
    int    num;

    if (rootfld == NULL)  return -1;
    
    for (num = 0, fo = rootfld; fo; fo = fo->next)  {
	num += fo->num;
    }
    return num;
}


Folder *folder_open(char *mailbox, int beh)
{
    Folder  *fo;
    Mail    *m;
    char    buf[SVRBUF], *idx[PARMAX], *pipeto;
    int     i, rs;

    if (rootfld == NULL)  {
	umask(0077);
	if (mktmpdir(gdir) == NULL)
	    return NULL;
    }	
    
    if (mailbox == NULL)  return NULL;

    if ((pipeto = rc_check(RCKEY_SIG)) != NULL)  {
	load_signature(pipeto);
    }

    pipeto = rc_check(RCKEY_PIPE);
    rs = ziptoken(mailbox, idx, PARMAX, DELIMS);
    for (i = 0; i < rs; i++)  {
	fullpath(idx[i], buf);
//	printf("folder: %s\n", buf);
	if (check_loaded(buf))  continue;  /* message folder loaded already */
	
	if ((fo = msg_loader(gdir, buf, fid, 1, beh)) != NULL)  {
	    logPrintf(LOG_INFO, "[%s] loaded, %d messagesd\n", buf, fo->num);
	    folder_insert(fo);
	    fid++;
	    
	    for (m = fo->mail; m; m = m->next)  {
		//printf("pipe message: %s (%X)\n", pipeto, m->behave);
		msg_pipeto(m, pipeto, 1, signature);
	    }
	}
    }

    return rootfld;
}


static char blkbuf[BLKBUF];

static int folder_backup(char *backup)
{
    Folder *fo;
    Mail   *m;
    FILE   *fin, *fout;
    int    cnt, rs;

    if (!backup)  return -1;
    
    fullpath(backup, blkbuf);
    fout = fopen(blkbuf, "a");
    if (fout == NULL)  {
	logPrintf(LOG_WARN, "folder_backup: open [%s] failed.\n", blkbuf);
	return -2;
    }
	
    for (cnt = 0, fo = rootfld; fo; fo = fo->next)  {
	for (m = fo->mail; m; m = m->next)  {
	    if (m->removed == 0)  continue;
	    
	    /* pick up removed messages, then backup them */
	    
	    if ((fin = fopen(m->msg->name, "r")) == NULL)  {
		logPrintf(LOG_WARN, "folder_backup: open [%s] failed.\n", 
				m->msg->name);
		continue;
	    }

	    while ((rs = fread(blkbuf, 1, BLKBUF, fin)) > 0)
		fwrite(blkbuf, 1, rs, fout);
	    fputs("\n", fout);
	    fclose(fin);
	    cnt++;
	}
    }
    logPrintf(LOG_INFO, "folder_backup: %d messages backuped.\n", cnt);
    return cnt;
}


static int folder_compress()
{
    Folder *fo;
    Mail   *m;
    int    cnt;

    for (cnt = 0, fo = rootfld; fo; fo = fo->next)  {

	/* unpipe undelivered messages */
	    
	for (m = fo->mail; m; m = m->next)  {
	    if (m->removed == 0)  {
		msg_unpipe(m);
		cnt++;
	    }
	}
	
	msg_combine(fo);
    }
    logPrintf(LOG_INFO, "folder_compress: %d messages undelivered.\n", cnt);
    return cnt;
}
	

int folder_close(char *backup, int beh)
{
    Folder *fo, *tmp;

    if (backup && (beh & ACT_BACK))  
	folder_backup(backup);
    
    if (beh & ACT_DELE)  
	folder_compress();
    
    fo = rootfld;
    while (fo)  {
	msg_clean(fo);
	tmp = fo;
	fo = fo->next;
	free(tmp);
    }

    rmdir(gdir);
    rootfld = NULL;
    return OK;
}


static char *ddmsg = "dead.letter";

Folder *folder_online(char *to, char *cc, char *bcc, char *sub, int beh)
{
    Folder  *fo;
    Mail    *m;
    FILE    *fp;
    char    buf[SVRBUF], *p;
    //int     rs;

    if (rootfld == NULL)  {
	umask(0077);
	if (mktmpdir(gdir) == NULL)
	    return NULL;
    }

    if ((fo = msg_loader(gdir, ddmsg, fid, 1, beh)) == NULL)  
	return NULL;
    
    folder_insert(fo);
    m = msg_expand(fo, gdir, NULL);

    p = rc_check(RCKEY_REPLY);
    if (p == NULL)  {
	sprintf(buf, "%s@%s", get_user(), get_hostname());
	p = buf;
	logPrintf(LOG_WARN, "folder_online: system address %s\n", p);
    }
    m->from    = dup_str(p);
    m->reply   = dup_str(p);
    m->to      = dup_str(to);
    m->cc      = dup_str(cc);
    m->bcc     = dup_str(bcc);
    m->subject = dup_str(sub);
    m->behave  = beh;
    
    fp = fopen(m->msg->name, "w");
    if (fp == NULL)  {
	msg_free_mailobj(m);
	msg_clean(fo);
	rmdir(gdir);
	rootfld = NULL;
	return NULL;
    }
    
    fprintf(fp, "From %s %s\n", email_pure(m->from), timenow());
    
    if (isatty(0))  {
    	if (sub == NULL)  {
    	    fprintf(stderr, "Subject: ");
    	    fgets(buf, SVRBUF, stdin);
    	    chop(buf);
    	    if (!isblankline(buf))  m->subject = dup_str(buf);
    	}
    	fprintf(stderr, "Use CTRL-D to terminate letter\n");
    }
    
    fprintf(fp, "From: %s\n", m->reply);
    if (m->to)  fprintf(fp, "To: %s\n", m->to);
    if (m->cc)  fprintf(fp, "Cc: %s\n", m->cc);
    if (m->subject)  fprintf(fp, "Subject: %s\n", m->subject);
   
    /*if ((m->behave & ACT_EHEAD) != 0)  {
	rs = 0;
	for (p = rc_check(RCKEY_HEADER); p; p = rc_check_next())  {
	    if (head_cmp(p, HD_MAILER))  rs++;
	    fprintf(fp, "%s\n", p);
	}
	if (rs == 0)  fprintf(fp, XMAILER, VERSION);
    }*/
    fprintf(fp, "\n");
		
    m->msg->boff = ftell(fp);
    while (fgets(buf, SVRBUF, stdin))  {
	if (!strncmp(buf, "From ", 5))  {
	    m->msg->len += fprintf(fp, ">%s", buf);
	} else {
	    m->msg->len += fprintf(fp, "%s", buf);
	}
	m->msg->line++;
    }

    if ((m->behave & ACT_ASIG) && signature)  {
	m->msg->len += fprintf(fp, "%s", signature);
    }

    fclose(fp);
    m->msg->mtime = filedate(m->msg->name, &m->msg->msize);

    if ((p = rc_check(RCKEY_PIPE)) != NULL)  msg_pipeto(m, p, 1, signature);
    /*if (p && (m->behave & ACT_PIPE))  
	msg_pipe2(m, p, signature);*/
    
    return fo;
}


int folder_close_online(char *backup, int beh)
{
    int  rs = rootfld->num;
    
    if (backup && (beh & ACT_BACK))  
	folder_backup(backup);
    
    if (beh & ACT_DELE)  
	rs = folder_compress();
    
    msg_clean(rootfld);
    free(rootfld);
    rootfld = NULL;

    rmdir(gdir);
    if (rs == 0)
	unlink(ddmsg);
    
    return OK;
}


void folder_dump(Folder *fld)
{
    Folder *fo;
    Mail   *m;
    FILE   *fp;
    char   buf[SVRBUF];

    for (fo = fld; fo; fo = fo->next)  {
	for (m = fo->mail; m; m = m->next)  {	
	    fp = fopen(m->msg->name, "r");
	    if (fp == NULL)  continue;
	    
	    while (fgets(buf, SVRBUF, fp))  {
		fputs(buf, stdout);
	    }
	    
	    fclose(fp);
	    printf("\n");
	}
    }
}


void folder_list(Folder *fld)
{
    struct winsize ws;
    Folder *fo;
    Mail   *m;
    char   buf[SVRBUF], ad[80];
    int    i, addr;

    if (ioctl(1, TIOCGWINSZ, (char *) &ws) < 0)  {
	ws.ws_col = atoi(getenv("COLUMNS"));
	ws.ws_row = atoi(getenv("LINES"));
	if (ws.ws_col == 0)  ws.ws_col = 80;
    }
    addr = ws.ws_col / 3;    /* 1/3 column for target address */
	
    for (fo = fld; fo; fo = fo->next)  {
	    
	printf("[%s]: %d messages.\n", fo->name, fo->num);
	
	for (m = fo->mail; m; m = m->next)  {
		
	    i = sprintf(ad, "%s", m->to ? m->to : "");
	    for ( ; i < addr; i++)  ad[i] = ' ';
	    ad[addr] = 0;
	    
	    i = sprintf(buf, "%5d# %-s %6d \"%s\"", m->id, ad, m->msg->len, 
			    m->subject ? m->subject : "");
	    
	    if (i >= ws.ws_col)  {
		buf[ws.ws_col - 2] = '"';
		buf[ws.ws_col - 1] = 0;
	    }
	    printf("%s\n", buf);
	}
    }	    
}


