
/*
   address.c  -- routines for handle email address. 
   
   Copyright (C) 1998, 1999  Xuming <xuming@iname.com>
   
   This program is free software; you can redistribute it and/or 
   modify it under the terms of the GNU General Public License as 
   published by the Free Software Foundation; either version 2, or 
   (at your option) any later version.
	   
   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of 
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License, the file COPYING in this directory, for
   more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/*
   These lexical expresses are picked from RFC822. I range them with 
   my view point, classified in order.

   address        =  mailbox / group
   ------------------------------------------------------------------
   group          =  phrase ":" [#mailbox] ";"
   mailbox        =  addr-spec / phrase route-addr
   ------------------------------------------------------------------
   addr-spec      =  local-part "@" domain
   phrase         =  1*word   
   route-addr     =  "<" [route] addr-spec ">"
   ------------------------------------------------------------------
   domain         =  sub-domain *("." sub-domain)
   local-part     =  word *("." word)   
   route          =  1#("@" domain) ":" 
   word           =  atom / quoted-string
   ------------------------------------------------------------------
   sub-domain     =  domain-ref / domain-literal
   atom           =  1*<any CHAR except specials, SPACE and CTLs>
   quoted-string  =  <"> *(qtext/quoted-pair) <">   ; Regular qtext or
                                                    ;   quoted chars.
   ------------------------------------------------------------------
   domain-literal =  "[" *(dtext / quoted-pair) "]"
   domain-ref     =  atom 
   qtext          =  <any CHAR excepting <">,     ; => may be folded
                     "\" & CR, and including linear-white-space>
   quoted-pair    =  "\" CHAR
   ------------------------------------------------------------------
   dtext          =  <any CHAR excluding "[",     ; => may be folded
                     "]", "\" & CR, & including linear-white-space>

*/
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <string.h>
#include "address.h"


#define	SMBUF	4096	/* should be enough for handling one address */

#define	islwsp(s)	(((s) > 0) && ((s) <= ' '))

enum { EA_START, EA_ADDPH, EA_ROUADDR, EA_ROUTE, EA_LCPT, 
	EA_DOMAIN, EA_DOMAIN2, EA_RTEND, EA_DIE };
    
static Iadr _iadr;

int ea_err;

static char *lex_parentheses(char *p, char **out);
static char *lex_quotext(char *p, char **out);
static char *lex_square(char *p, char **out);
static void *sm_malloc(int size);
static void sm_free();

#ifndef	HAVE_STRCASECMP
static int  strcasecmp(char *sour, char *dest);
#endif


#ifdef	EXECUTABLE

int main()
{
    char  buf[1024], *p;
    char  *unin = "none";
    Iadr  *iadr;

    while (fgets(buf, 1024, stdin))  {
	p = buf;
	while ((iadr = email_addr(p)) != NULL)  {
	    printf("email:  %s\n", iadr->email ? iadr->email : unin);
	    printf("local:  %s\n", iadr->local ? iadr->local : unin);
	    printf("domain: %s\n", iadr->domain ? iadr->domain : unin);
	    printf("phrase: %s\n", iadr->phrase ? iadr->phrase : unin);
	    printf("route:  %s\n", iadr->route ? iadr->route : unin);
	    printf("rtto:   %s\n\n", iadr->rtto ? iadr->rtto : unin);
	    if ((p = iadr->next) == NULL)  break;
	}
    }
    return 0;	    
}

#endif



/* extract host specified email address from complexical Internet address.
   e.g: "XUming" <xuming@iname.com>   --> xuming@iname.com

   For making the life easier, I just choose the two simple branches. The
   group address might append later, and do not take care of domain-literal:
   
   address = mailbox = addr-spec = local-part "@" domain
   = word *("." word) "@" atom *("." atom)
   address = mailbox = phrase route-addr = 1*word "<" [route] addr-spec ">"
   = 1*word "<" [1#("@" atom *("." atom)) ":"] 	\
   		word *("." word) "@" atom *("." atom) ">"

   status figure: 
   
   +----------->+-----------------","--------------------------->+
   |            ^                                                |
   |            |                                                |
 START -word-> ADDPH ----------"@"---------> DOMAIN ----","---> DIE
                |                                                ^
               "<"                                               |     
                |     +-------------">"------------->+           |
                |     |                              |           |
              ROUADDR --"@"--> ROUTE ------">"---> RTEND --","-->+
                |                |                   ^           |
               word             ":"                  |           |
                |                |   +------------->">"          |
                |                |   |               |           |
                +-------------> LCPT -"@"-> DOMAIN2 -+-----","-->+
                                 |                               |
                                 +--------------","--------------+
				 
   Note any time a ASCII0 should push the status to DIE
*/

Iadr *email_addr(char *iaddr)
{
    char  *buf, *p, *k, *tmp;
    int   len, qflag;

    sm_free();
    memset(&_iadr, 0, sizeof(Iadr));
    p = buf = sm_malloc(1024);	/* hope its enough */
    qflag = EA_START;
    k = NULL;
    ea_err = 0;	/* ok */
    
    while (iaddr) {
	switch (qflag) {
	case EA_START:
	    switch (*iaddr) {
	    case ',':
	    case '\0':
		qflag = EA_DIE;
		break;
		
	    default:
		if (islwsp(*iaddr))  iaddr++;
		else qflag = EA_ADDPH;
		break;
	    }
	    break;
	    
	case EA_ADDPH:
	    switch (*iaddr) {
	    case '"':
		iaddr = lex_quotext(iaddr, &p);
		break;

	    case '(':
		iaddr = lex_parentheses(iaddr, NULL);
		break;

	    case '\n':
	    case '\t':
	    case '\f':
	    case '\r':
	    case '\v':
		*p++ = ' ', iaddr++;
		break;

	    case '@':
		qflag = EA_DOMAIN;
		*p = '\0';
		if ((_iadr.local = sm_malloc(strlen(buf) + 1)) != NULL)
		    strcpy(_iadr.local, buf);
		p = buf;
		iaddr++;
		break;
		
	    case '<':
		qflag = EA_ROUADDR;
		*p = '\0';
		if ((_iadr.phrase = sm_malloc(strlen(buf) + 1)) != NULL)
		     strcpy(_iadr.phrase, buf);
		p = buf;
		*p++ = *iaddr++;	/* save first '<' */
		break;
		
	    case ',':
	    case '\0':
		qflag = EA_DIE;
		*p = '\0';
		if ((_iadr.local = sm_malloc(strlen(buf) + 1)) != NULL)
		    strcpy(_iadr.local, buf);
		break;

	    default:
		*p++ = *iaddr++;
	    	break;
	    }
	    break;
	    
	case EA_ROUADDR:
	    switch (*iaddr) {
	    case '>':
		qflag = EA_RTEND;
		*p++ = *iaddr++, *p = '\0';
		if ((_iadr.rtto = sm_malloc(strlen(buf) + 1)) != NULL)
		    strcpy(_iadr.rtto, buf);
		break;
		
	    case '@':
		qflag = EA_ROUTE;
		k = p;		/* remember this place, starting of ROUTE */
		break;
		
	    case '\0':
		qflag = EA_DIE;
		ea_err = 1;		/* syntax error */
		break;

	    default:
	    	if (islwsp(*iaddr))  iaddr++;
		else {
		    qflag = EA_LCPT;
		    k = p;
		}
		break;
	    }
	    break;
		    
	case EA_ROUTE:
	    switch (*iaddr) {
	    case '"':  
		iaddr = lex_quotext(iaddr, &p);
		break;
		
	    case '(':
		iaddr = lex_parentheses(iaddr, NULL);
		break;
		
	    case '\0':
		qflag = EA_DIE;
		ea_err = 2;		/* syntax error */
		break;
		
	    case '>':
		qflag = EA_RTEND;
		*p++ = *iaddr++, *p = '\0';
		if ((_iadr.rtto = sm_malloc(strlen(buf) + 1)) != NULL)
		    strcpy(_iadr.rtto, buf);
		len = p - k + 2;
		if ((_iadr.route = sm_malloc(len)) != NULL)  {
		    tmp = _iadr.route;
		    while (*k != '>')  *tmp++ = *k++;
		    *tmp = '\0';
		}
		break;
		
	    case ':':
		qflag = EA_LCPT;
		*p++ = *iaddr++;
		len = p - k + 2;
		if ((_iadr.route = sm_malloc(len)) != NULL)  {
		    tmp = _iadr.route;
		    while (*k != ':')  *tmp++ = *k++;
		    *tmp++ = *k++, *tmp = '\0';
		}
		break;
		
	    default:
		if (islwsp(*iaddr)) iaddr++;
		else *p++ = *iaddr++;
		break;
	    }
	    break;
		    
	case EA_LCPT:
	    switch (*iaddr) {
	    case '"':
		iaddr = lex_quotext(iaddr, &p);
		break;
		
	    case '(':
		iaddr = lex_parentheses(iaddr, NULL);
		break;
		
	    case '\0':
		qflag = EA_DIE;
		ea_err = 3;
		break;
		
	    case '>':
		qflag = EA_RTEND;
		*p++ = *iaddr++, *p = '\0';
		if ((_iadr.rtto = sm_malloc(strlen(buf) + 1)) != NULL)
		    strcpy(_iadr.rtto, buf);
		if (k != NULL)  {
		    len = p - k + 2;
		    if ((_iadr.local = sm_malloc(len)) != NULL)  {
			tmp = _iadr.local;
			while (*k != '>') *tmp++ = *k++;
			*tmp++ = '\0';
		    }
		}
		break;

	    case ',':	  /* it's syntax error in fact */
		qflag = EA_DIE;
		*p = '\0';
		if ((_iadr.rtto = sm_malloc(strlen(buf) + 1)) != NULL)
		    strcpy(_iadr.rtto, buf);
		if (k != NULL)  {
		    len = p - k + 2;
		    if ((_iadr.local = sm_malloc(len)) != NULL)  {
			tmp = _iadr.local;
			while (*k) *tmp++ = *k++;
			*tmp++ = '\0';
		    }
		}
		break;
		
	    case '@':
		qflag = EA_DOMAIN2;
		*p++ = *iaddr++;
		if (k != NULL)  {
		    len = p - k + 2;
		    if ((_iadr.local = sm_malloc(len)) != NULL)  {
			tmp = _iadr.local;
			while (*k != '@') *tmp++ = *k++;
			*tmp++ = '\0', k++;
		    }
		}
		break;

	    default:
		if (islwsp(*iaddr))  iaddr++;
		else *p++ = *iaddr++;
		break;
	    }
	    break;
	    
	case EA_DOMAIN:
	    switch (*iaddr) {
	    case '"':
		iaddr = lex_quotext(iaddr, &p);
		break;
		
	    case '(':
		iaddr = lex_parentheses(iaddr, NULL);
		break;
		
	    case ',':
	    case '\0':
	    case '>':	/* this should be syntax error */
		qflag = EA_DIE;
		*p = '\0';
		if ((_iadr.domain = sm_malloc(strlen(buf) + 1)) != NULL)
		    strcpy(_iadr.domain, buf);
		break;

	    default:
		if (islwsp(*iaddr))  iaddr++;
		else *p++ = *iaddr++;
		break;
	    }
	    break;
	    
	case EA_DOMAIN2:
	    switch (*iaddr) {
	    case '"':
		iaddr = lex_quotext(iaddr, &p);
		break;
		
	    case '(':
		iaddr = lex_parentheses(iaddr, NULL);
		break;
		
	    case ',':
	    case '\0':	/* it's syntax error in fact */
		*p = '\0';
		qflag = EA_DIE;
		if ((_iadr.rtto = sm_malloc(strlen(buf) + 1)) != NULL)
		    strcpy(_iadr.rtto, buf);
		if (k != NULL)  {
		    len = p - k + 2;
		    if ((_iadr.domain = sm_malloc(len)) != NULL)  {
			tmp = _iadr.domain;
			while (*k)  *tmp++ = *k++;
			*tmp = '\0';
		    }
		}
		break;
		
	    case '>':
		*p++ = *iaddr++, *p = '\0';
		qflag = EA_RTEND;
		if ((_iadr.rtto = sm_malloc(strlen(buf) + 1)) != NULL)
		    strcpy(_iadr.rtto, buf);
		if (k != NULL)  {
		    len = p - k + 2;
		    if ((_iadr.domain = sm_malloc(len)) != NULL)  {
			tmp = _iadr.domain;
			while (*k != '>')  *tmp++ = *k++;
			*tmp = '\0';
		    }
		}
		break;

	    default:
		if (islwsp(*iaddr))  iaddr++;
		else *p++ = *iaddr++;
		break;
	    }
	    break;
		
	case EA_RTEND:
	    if ((*iaddr == ',') || !*iaddr)  qflag = EA_DIE;
	    else iaddr++;
	    break;
	    
	case EA_DIE:
	default:
	    break;
	}
	
	if (qflag == EA_DIE)  {
	    if (*iaddr == ',')  _iadr.next = iaddr + 1;
	    break;
	}
    }

    if (ea_err)  return NULL;
    
    len = 0;
    if (_iadr.local)  len = strlen(_iadr.local);
    if (_iadr.domain) len += strlen(_iadr.domain);
    if ((_iadr.email = sm_malloc(len + 4)) != NULL)  {
	if (_iadr.local)  strcpy(_iadr.email, _iadr.local);
	if (_iadr.domain) {
	    strcat(_iadr.email, "@");
	    strcat(_iadr.email, _iadr.domain);
	}
    }
    return &_iadr;
}

int email_cmp(char *sour, char *dest)
{
    char  tmp[512];
    Iadr  *iadr;

    if ((iadr = email_addr(sour)) == NULL)  {
	if ((iadr = email_addr(dest)) == NULL)  return 0;
	else  return 1;
    }
    strncpy(tmp, iadr->email, 511);
    tmp[511] = '\0';

    if ((iadr = email_addr(dest)) == NULL)  return 1;
    return strcasecmp(tmp, iadr->email);
}

/* get pure address, composed by local@domain  */

char *email_pure(char *addr)
{
    Iadr  *iadr;

    if ((iadr = email_addr(addr)) == NULL)  return NULL;
    return iadr->email;
}

/* if there is routed address (quoted by angle brackets "<>"), 
   take it first. Or return pure address */ 

char *email_route(char *addr)
{
    Iadr  *iadr;

    if ((iadr = email_addr(addr)) == NULL)  return NULL;
    if (iadr->rtto)  return iadr->rtto;
    return iadr->email;
}

/* comment */

static char *lex_parentheses(char *p, char **out)
{
    if (out)  *(*out)++ = *p++; 
    else p++;    /* output the starting parenthes */
    
    while (*p) {
	switch (*p)  {
	case '\\':
	    p++;
	    if (*p)  {
		if (out)  *(*out)++ = *p++;
		else p++;
	    }
	    break;

	case '(':
	    p = lex_parentheses(p, out);
	    break;

	case ')':
	    if (out)  *(*out)++ = *p++;
	    else p++;
	    return p;
	    
	default:
	    if (out)  *(*out)++ = *p++;
	    else p++;
	    break;
	}
    }
    return p;
}

/* quoted-string */

static char *lex_quotext(char *p, char **out)
{
    if (out)  *(*out)++ = *p++;      
    else  p++;	/* output the starting quotation */

    while (*p) {
	switch (*p)  {
	case '\\':
	    p++;
	    if (*p)  {
		if (out)  *(*out)++ = *p++;
		else p++;
	    }
	    break;

	case '(':
	    p = lex_parentheses(p, out);
	    break;

	case '"':
	    if (out)  *(*out)++ = *p++;
	    else p++;
	    return p;
	    
	default:
	    if (out)  *(*out)++ = *p++;
	    else p++;
	    break;
	}
    }
    return p;
}
	
/* domain-literal, which the appropriate name-domain is to use directly,
   bypassing normal name-resolution mechanisms.*/

static char *lex_square(char *p, char **out)
{
    if (out == NULL)  p++;
    else  *(*out)++ = *p++;      /* output the starting square bracket */

    while (*p) {
	switch (*p)  {
	case '\\':
	    p++;
	    if (*p)  {
		if (out)  *(*out)++ = *p++;
		else p++;
	    }
	    break;

	case '(':
	    p = lex_parentheses(p, out);
	    break;

	case '"':
	    p = lex_quotext(p, out);
	    break;

	case '[':
	    p = lex_square(p, out);
	    break;

	case ']':
	    if (out)  *(*out)++ = *p++;
	    else p++;
	    return p;
	    
	default:
	    if (out)  *(*out)++ = *p++;
	    else p++;
	    break;
	}
    }
    return p;
}


#ifndef	HAVE_STRCASECMP

static int strcasecmp(char *sour, char *dest)
{
    char    a, b;

    while (1)  {
        if ((*sour < 'a') || (*sour > 'z'))  a = *sour;
        else  a = *sour - 'a' + 'A';

        if ((*dest < 'a') || (*dest > 'z'))  b = *dest;
        else  b = *dest - 'a' + 'A';

        if (a != b)  return 1;
        if (a == '\0')  break;
        sour++, dest++;
    }
    return 0;
}

#endif


/* these routines is used to allocate tiny memory for such as short
   string etc. Usaul routines like malloc() take too much sources. */

static char _smbuf[SMBUF];
static int  _smroom;
static char *_smcur;

static void *sm_malloc(int size)
{
    char *tmp = _smcur;
    
    if (_smroom < size)  return NULL;
    _smcur  += size;
    _smroom -= size;
    return tmp;
}

static void sm_free()
{
    _smroom = SMBUF;
    _smcur  = _smbuf;
}

