/*
   debug.c  -- handle debug and log method. 
   
   This module exports some procedures about log file and debug 
   information output. 
   
   Copyright (C) 1998, 1999  Xuming <xuming@iname.com>
   
   This program is free software; you can redistribute it and/or 
   modify it under the terms of the GNU General Public License as 
   published by the Free Software Foundation; either version 2, or 
   (at your option) any later version.
	   
   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of 
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License, the file COPYING in this directory, for
   more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <sys/stat.h>
#include <fcntl.h>

#if HAVE_UNISTD_H
  #include <sys/types.h>
  #include <unistd.h>
#endif

#if TIME_WITH_SYS_TIME
  #include <sys/time.h>
  #include <time.h>
#else
  #if HAVE_SYS_TIME_H
    #include <sys/time.h>
  #else
    #include <time.h>
  #endif
#endif

#define	DBGBUF	2048
#define STDOUT	1
#define LOGLVL	0


static int  logfd = STDOUT;	/* set default to standard out */

static char _tmsp[32];
static char *_mon[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
    		 	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

static char dbgbuf[DBGBUF];


/* get the current time stamp */
			
static char *time_stamp()
{
    struct tm *t;
    time_t sec;

    time(&sec);
    t = localtime(&sec);
    sprintf(_tmsp, "%s %2d %02d:%02d:%02d", _mon[t->tm_mon], 
    		t->tm_mday, t->tm_hour, t->tm_min, t->tm_sec);
    return _tmsp;
}


/* open the specified log device or file.
   log: name of log device or file.
   quota: if log into file, it's size limitation, 0 means unlimit.
   return: the file description of log file or device */

int  log_open(char *log, int quota)
{
    struct stat	fs;
    
    if ((logfd != STDOUT) || !log)  {
#ifdef DEBUG
	printf("log device has opened already.\n");
#endif
	return logfd;
    }

    stat(log, &fs);
    if (S_ISREG(fs.st_mode) && quota && (fs.st_size > quota))  
	truncate(log, 0);
    
    logfd = open(log, O_WRONLY | O_NOCTTY | O_APPEND | O_CREAT,
		    S_IRUSR | S_IWUSR);
    if (logfd < 0)  {
	perror("open log");
	logfd = STDOUT;
    }
    return logfd;
}

void log_close()
{
    if (logfd != STDOUT)  close(logfd);
    logfd = STDOUT;
}


/* just logging formated string. If no logging device is specified
   the information is discarded. */
    
int logPrintf(int lvl, char *fmt, ... )
{
    int	  len;

#if	defined(va_start)
    va_list	ap;
#endif
    
    if ((lvl < LOGLVL) || (logfd == STDOUT))  return 0;
    
    len = sprintf(dbgbuf, "%s ", time_stamp());
    
#if	defined(va_start)
    
    va_start(ap, fmt);
    len += vsnprintf(dbgbuf + len, DBGBUF - len, fmt, ap);
    va_end(ap);
    
#else
    
    len += vsnprintf(dbgbuf + len, DBGBUF - len, fmt, (void*)(&fmt + 1));
    
#endif
    
    len = write(logfd, dbgbuf, len);
    return len;
}


/* write formated string to both logging devices and console */

int dbgPrintf(char *fmt, ... )
{
    int	  len;
    
#if	defined(va_start)
    va_list	ap;
#endif
    
    len = sprintf(dbgbuf, "%s ", time_stamp());
    
#if	defined(va_start)
    
    va_start(ap, fmt);
    len += vsnprintf(dbgbuf + len, DBGBUF - len, fmt, ap);
    va_end(ap);
    
#else
    
    len += vsnprintf(dbgbuf + len, DBGBUF - len, fmt, (void*)(&fmt + 1));
   
#endif
    
    len = write(2, dbgbuf, len);

    if (logfd != STDOUT)  len = write(logfd, dbgbuf, len);
    
    return len;
}

    
/* directly print strings to standard out without buffer, no logging */

int conPrintf(char *fmt, ...)
{
    int   len;
    
#if	defined(va_start)
    va_list	ap;
#endif
    
#if	defined(va_start)
    
    va_start(ap, fmt);
    len = vsnprintf(dbgbuf, DBGBUF, fmt, ap);
    va_end(ap);

#else
   
    len = vsnprintf(dbgbuf, DBGBUF, fmt, (void*)(&fmt + 1));

#endif
    
    return write(1, dbgbuf, len);
}

/* directly print strings to standard error without buffer, no logging */

int errPrintf(char *fmt, ...)
{
    int   len;
    
#if	defined(va_start)
    va_list	ap;
#endif
    
#if	defined(va_start)
    
    va_start(ap, fmt);
    len = vsnprintf(dbgbuf, DBGBUF, fmt, ap);
    va_end(ap);

#else
   
    len = vsnprintf(dbgbuf, DBGBUF, fmt, (void*)(&fmt + 1));

#endif
    
    return write(2, dbgbuf, len);
}

