
/*
   parse.c -- general routine for parse user option file
	
   Copyright (C) 1998, 2000  Xuming <xuming@iname.com>
   
   This program is free software; you can redistribute it and/or 
   modify it under the terms of the GNU General Public License as 
   published by the Free Software Foundation; either version 2, or 
   (at your option) any later version.
	   
   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of 
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License, the file COPYING in this directory, for
   more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/
/*
   This file exposes some routine for parsing resource configuration file.
   It stores all configuration items formed as keys with a body, linke this:
   
   +---+          +-------+-----------------+
   | c |--------> |  key  |   body          |
   +---+          +-------+-----------------+
   | a |---+      +-------+-----------------+
   +---+   |----> |  key  |   body          |
   | c |          +-------+-----------------+
   +---+
   | h |             ...
   +---+
   | e |
   +---+

   Any configuration items must composed by a key and one line or several 
   lines body, e.g:

   default_action  pop3, smtp

   There the key is 'default_action' and other contents are of body. They
   are seperated by space character. The body can cross several lines with
   this form:

   default_action {
       pop3
       smtp
   }

   These two configurater are stored different body. First one saves
   'pop3, smtp', and another one saves this: '{#pop3#smtp#}'. So you must
    handle it by your self.
*/ 

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>


/* size of a parsing cache, the MAXIMEM capability of RC file */
#define PCACHE	1024	

/* maximem size of a RC key */
#define KEYMAX	128

/* MAXIMEM size of a single item */
#define SITEM	4096	


static int	commt = '#';
static char	*cache [PCACHE];
static int	clen = 0;		/* available cache item */

static char	skey[KEYMAX];
static int	sidx = 0;


/* compress the inputing line, get rid of the comment and spare space
   at the beginning and ending parts. e.g:

   "   hello   # this is hello world " --> "hello"
*/

static char *line_refine(char *s, int commt)
{
    char  *p;

#ifdef	DEBUG
    if (!s)  return NULL;
#endif	

    /* search the comment start */

    p = s;
    while ((p = strchr(p, commt)) != NULL)  {
        if ((p != s) && (*(p-1) == '\\'))  continue;
        *p = '\0';
        break;
    }

    /* shorten the tail */

    p = s + strlen(s);
    while ((p != s) && (*p >= 0) && (*p <= ' '))  *p-- = '\0';

    /* compress the head */

    for (p = s; isspace(*p); p++);
    return p;
}

#ifndef	HAVE_STRCASECMP

static int strcasecmp(char *sour, char *dest)
{
    char    a, b;

    while (1)  {
	if ((*sour < 'a') || (*sour > 'z'))  a = *sour;
	else  a = *sour - 'a' + 'A';
	
        if ((*dest < 'a') || (*dest > 'z'))  b = *dest;
	else  b = *dest - 'a' + 'A';

	if (a != b)  return 1;
	if (a == '\0')  break;
	sour++, dest++;
    }
    return 0;
}

#endif

/* The key and body are ready. Then insert them into configuration cache */

static char *add_cache(char *key, char *body, char *ext)
{
    int  len;

#ifdef	DEBUG
    if (!key || !body || !ext)  return NULL;
#endif
    
    if (clen == PCACHE) return NULL;

    len = strlen(key);
    cache[clen] = (char*) malloc(len + strlen(body) + strlen(ext) +  3);
    if (cache[clen] == NULL)  return NULL;

    strcpy(cache[clen], key);
    strcpy(cache[clen] + len + 1, body);
    strcpy(cache[clen] + len + strlen(body) + 2, ext);
    clen++;
    
    return cache[clen-1];
}

/* collect crossing line configure, e.g:

   default_action {
       pop3
       smtp
   }

   make result in the gbuf is "{#pop3#smtp#}".
   Note that the size of buffer is limited by SITEM, so do NOT make 
   too larger configure. */

static char *get_complex(FILE *fp, char *gbuf)
{
    char  buf[512], *idx;
    int   off;

    off = sprintf(gbuf, "{#");
    while (fgets(buf, 512, fp))  {
	
	idx = line_refine(buf, commt);
	if (*idx == '\0')  continue;
	
	if (!strcmp(idx, "}"))  break;
	if (off + strlen(idx) + 1 >= SITEM)  break;
	
	off += sprintf(gbuf + off, "%s#", idx);
    }
    sprintf(gbuf + off, "}");
    return gbuf;
}


int rc_open(char *s)
{
    FILE  *fp;
    char  buf[512], *p, *key, *body, *gbuf;

    if ((fp = fopen(s, "r")) == NULL)  return -1;

    if ((gbuf = (char*) malloc(SITEM)) == NULL) return -1;
    
    while (fgets(buf, 512, fp))  {

	key = line_refine(buf, commt);
	if (*key == '\0')  continue;

	for (body = key; *body && !isspace(*body); body++);
	if (*body)  *body++ = '\0';

	while (*body && isspace(*body))  body++;
	
	if (*(p = body) != '\0')  p += strlen(p) - 1;
	
	if (*p == '{')  *p = '\0', get_complex(fp, gbuf);
	else  gbuf[0] = '\0';
	add_cache(key, body, gbuf);
    }
    *skey = '\0';
    sidx = clen;

    free(gbuf);
    fclose(fp);
    return clen;
}


void rc_close()
{
    int i;

    for (i = 0; i < clen; i++)  {
	free(cache[i]);
	cache[i] = NULL;
    }
    clen = 0;
    *skey = '\0';
    sidx = clen;
}

char *rc_check(char *key)
{
    int  i;

    for (i = 0; i < clen; i++)  {
	if (!strcasecmp(key, cache[i]))  {
	
	    /* storing last key and the successful point. It's ready for 
	       checking next item. */
		
	    sidx = i;
	    strncpy(skey, key, KEYMAX - 1);
	    skey[KEYMAX - 1] = '\0';

	    return cache[i] + strlen(key) + 1;
	}
    }
    return NULL;
}


char *rc_check_next()
{
    for (sidx++; sidx < clen; sidx++)  {
	if (!strcasecmp(skey, cache[sidx]))  {
	    return cache[sidx] + strlen(skey) + 1;
	}
    }
    return NULL;
}

void rc_comment(int ch)
{
    commt = ch;
}
		

#ifdef	EXECUTABLE

int main(int argc, char **argv)
{
    int i, len;
    char buf[256], *p;
    
    if (argc < 2)  {
	printf("usage: %s rc_file\n", argv[0]);
	return 0;
    }

    len = rc_open(argv[1]);
    if (len < 0)  return len;

    for (i = 0; i < len; i++)  {
	printf("%s\n", cache[i]);
	p = cache[i] + strlen(cache[i]) + 1;
	printf("  %s\n", p);
	p += strlen(p) + 1;
	printf("  %s\n", p);
    }

    while (fgets(buf, 256, stdin)) {
	buf [strlen(buf) - 1] = '\0';
	if ((p = rc_check(buf)) == NULL)  continue;

	printf("-> %s\n", p);
	p += strlen(p) + 1;
	printf("   %s\n", p);
	while ((p = rc_check_next()) != NULL)  {
	    printf("-> %s\n", p);
	    p += strlen(p) + 1;
	    printf("   %s\n", p);
	}
    }
	
    rc_close();
    return 0;
}

#endif

