
/*
   pmalloc.c -- routines for allocate scramble memory
	
   Copyright (C) 1998, 1999  Xuming <xuming@iname.com>
   
   This program is free software; you can redistribute it and/or 
   modify it under the terms of the GNU General Public License as 
   published by the Free Software Foundation; either version 2, or 
   (at your option) any later version.
	   
   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of 
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License, the file COPYING in this directory, for
   more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdlib.h>
#include <string.h>
#include "pmalloc.h"


static PMAL *pmal = NULL;
int  pm_errno; 


PMAL *pm_handler(int items, int size)
{
    PMAL    *pm;
    size_t  len;

    pm_errno = PM_SUCCESS;
    
    len = sizeof(PMAL) + items * sizeof(void*) + size;
    if ((pm = malloc(len)) == NULL)  {
	pm_errno = PM_NOMEM;
	return NULL;
    }

    memset(pm, 0, len);
    pm->items = items;
    pm->space = size;
    
    pm->item = (char**) & pm->mem;
    pm->cur = pm->mem + pm->items * sizeof(void*);
    
    if (pmal == NULL)  pmal = pm;
    else  {
	pm->next = pmal;
	pmal->prev = pm;
    }
    return pm;
}

void pm_close(PMAL *pm)
{
    PMAL  *tmp, *s;
    
    if (pm == NULL)  {
	pm_errno = PM_NULLOP;
	return;
    }
    
    if (pmal == pm)  pmal = pm->next;
    if (pm->prev) pm->prev->next = pm->next;
    if (pm->next) pm->next->prev = pm->prev;
    
    s = pm->sub;
    while (s)  {
	tmp = s;
	s = s->sub;
	free(tmp);
    }

    free(pm);
    pm_errno = PM_SUCCESS;
}

void pm_destruct()
{
    PMAL  *pm, *s, *tmp;

    pm = pmal;
    while (pm)  {

	s = pm->sub;
	while (s) {
	    tmp = s;
	    s = s->sub;
	    free(tmp);
	}
	
	tmp = pm;
	pm = pm->next;
	free(tmp);
    }
    pm_errno = PM_SUCCESS;
}


void *pm_alloc(PMAL *pm, int size)
{
    PMAL *sc;
    int  i = 0;
    
    if (pm == NULL)  {
	pm_errno = PM_NULLOP;
	return NULL;
    }
    
    if (size > pm->space)  {
	pm_errno = PM_LBLOCK;
	return NULL;
    }

    for (sc = pm; sc; sc = sc->sub)  {
	if (size > sc->space - sc->used)  continue;
	for (i = 0; i < sc->items; i++)  {
	    if (sc->item[i] == NULL)  break;
	}
	if (i == sc->items)  continue;

	break;
    }

    /* if no available room, then allocate one */
    
    pm_errno = PM_SUCCESS;
    if (sc == NULL)  {
	i = sizeof(PMAL) + pm->items * sizeof(void*) + pm->space;
	if ((sc = malloc(i)) == NULL)  {
	    pm_errno = PM_NOMEM;
	    return NULL;
	}

	memset(sc, 0, i);
	sc->items = pm->items;
	sc->space = pm->space;
    
	sc->item = (char**) & sc->mem;
	sc->cur  = sc->mem + sc->items * sizeof(void*);

	sc->sub = pm->sub;
	pm->sub = sc;
	
	i = 0;   /* start with with first item */
	pm_errno = PM_CONTINU;
    }
    
    sc->item[i] = sc->cur;
    sc->cur  += size;
    sc->used += size;
    
    return sc->item[i];
}


void pm_free(PMAL *pm)
{
    PMAL  *sc;
    
    if (pm == NULL)  {
	pm_errno = PM_NULLOP;
	return;
    }
    
    for (sc = pm; sc; sc = sc->sub)  {
	sc->used = 0;
	memset(sc->item, 0, sc->items * sizeof(void*));
	sc->cur = sc->mem + sc->items * sizeof(void*);
    }
    
    pm_errno = PM_SUCCESS;
}

char *pm_dup_str(PMAL *pm, char *str)
{
    char *p, *tmp;

    if (pm == NULL)  {
	pm_errno = PM_NULLOP;
	return NULL;
    }
    
    if (str == NULL)  {
	pm_errno = PM_NULLOP;
	return NULL;
    }
    
    p = tmp = pm_alloc(pm, strlen(str) + 1);
    if (p)  
	while ((*p++ = *str++) != '\0');
    return tmp;
}


#ifdef	EXECUTABLE

#include <stdio.h>

int main()
{
    PMAL *pm, *sc;
    char buf[256];
    int  i, k;

    pm = pm_handler(8, 128);
    pm_handler(8, 128);
    pm_close(pm);
    pm = pm_handler(4, 16);
    printf("create 3 PMAL objects.\n");
    printf("items = %d\n", pm->items);
    printf("space = %d\n", pm->space);
    printf("current = %p\n", pm->cur);
    for (i = 0; i < 2000; i += 100)  {
	sprintf(buf, "%d", i);
	pm_dup_str(pm, buf);
    }
    for (sc = pm; sc; sc = sc->sub)  {
	printf("used = %d ", sc->used);
	for (k = 0; k < sc->items; k++)  {
	    if (sc->item[k])  printf("[%s]", sc->item[k]);
	}
	printf("\n");
    }
    pm_free(pm);
    printf("current = %p\n", pm->cur);
    printf("used = %d\n", pm->used);
    return 0;
}

#endif
    
