
/*
   mailz.c -- a simple MUA for replacing mailx(1)
	
   Copyright (C) 1998, 2000  Xuming <xuming@iname.com>
   
   This program is free software; you can redistribute it and/or 
   modify it under the terms of the GNU General Public License as 
   published by the Free Software Foundation; either version 2, or 
   (at your option) any later version.
	   
   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of 
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License, the file COPYING in this directory, for
   more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <fcntl.h>
#include <signal.h>
#include <unistd.h>
#include <termios.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/ioctl.h>

#ifdef HAVE_READLINE_READLINE_H
  #include <readline/readline.h>
#else
  #ifdef HAVE_READLINE_H
    #include <readline.h>
  #endif
#endif
#ifdef HAVE_READLINE_HISTORY_H
  #include <readline/history.h>
#else
  #ifdef HAVE_HISTORY_H
    #include <history.h>
  #endif
#endif

#include "defs.h"
#include "memo.h"


#define SHOWMSG	"mailz.dump"


/* enviroment variables */

static Folder	*folder = NULL;
static Folder 	*fcur;		/* point to current folder */
static int	fid = 1;

static char	gdir[] = TMPPATH;
static int	lines = 25;
static int	cols = 80;

static struct tm  *today;

/* this two objects are used to DUMP messages MIME encoded */

static Head  *phead;
static char  *pdmsg;

/* options variables */

static int hidrm = 1;	/* 1: hide removed messages */

static int maxfld = 0;	/* 0: unlimit loading folders */

static int ps1 = '&';	/* prompt character */

static int ensig = 0;	/* 0: disable signature, 1: enable */

static int mime = 2;	/* 0: no handle to message, display the original */
			/* 1: MIME decode, resever full head */
			/* 2: MIME decode, part heads specified in "headmsk" */

static char *pager;
static char *editor;
static char *postpone;
static char *replier;
static char *addrmemo;		/* address memo */
static char *quota = DEFQUOTA;	/* quotation for replying message */
static char *signature = NULL;	/* signature file */
static char *headmsk = NULL;


/* check if the specified folder has been loaded. */

static Folder *check_loaded(char *fname)
{
    Folder *fo;

    for (fo = folder; fo; fo = fo->next)  {
	if (!strcmp(fo->name, fname))  
	    break;
    }
    return fo;
}


static void folder_statis()
{
    Folder  *fo;
    int     num, unrd, pp, ppfld, fld;

    num = unrd = pp = fld = ppfld = 0;
    for (fo = folder; fo; fo = fo->next)  {
	num += fo->num + fo->rm;
	unrd += fo->unrd;
	if (fo->pp) 
	    ppfld++, pp += fo->num + fo->rm;
	fld++;
    }
    printf("%d message folders loaded; total %d messages, %d unread.\n",
		    fld, num, unrd);
    if (pp)  
	printf("Including %d postponed messages within %d folders.\n", 
		    pp, ppfld);
}
		

/* get the number of loaded folder */

static int folder_num()
{
    Folder *fo;
    int  num;

    for (num = 0, fo = folder; fo; fo = fo->next, num++);
    return num;
}
	    
/* move current message to first unread message. */

static int next_unread(Folder *fo)
{
    Mail  *m;

    for (m = fo->mail; m; m = m->next)  {
	if (!m->oldmail) {	/* if message is unread... */
	    fo->cur = m->id;
	    /* fo->lst = m->id; */
	    break;
	}
    }
    return fo->cur;
}

/* load message folder.
 * parameter string composed by folder name seperated by space.
 * the 'pp' switch is for marking postpone class folder.
   The folder queue is constructed as double chain queue. The last prev 
   and next pointer point to null.
 */

static int loader(char *s, int pp)
{
    int    i, rs;
    char   *idx[PARMAX];
    char   *ftype[2] = { "Folder", "Postpone folder" };
    Folder *fo, *t;

    if (!s || !*s)  {
	printf("No folder specified.\n");
	return 0;
    }
    
    rs = ziptoken(s, idx, PARMAX, " ");
    for (i = 0; i < rs; i++)  {

	if (check_loaded(idx[i]))  {
	    printf("%s [%s] has loaded already.\n", ftype[pp], idx[i]);
	    continue;
	}
 
	if (maxfld && (folder_num() > maxfld))  
	    continue;
		
	if ((fo = msg_loader(gdir, idx[i], fid, pp, ACT_DEFT)) != NULL)  {
		
	#ifdef	DEBUG
	    printf("%s [%s] loaded, %d active messages %d unread\n", 
			ftype[pp], idx[i], fo->num, fo->unrd);
	#endif

	    /* insert this folder at the end of folder queue */
	    
	    if (folder == NULL)  {
		folder = fo;
		fo->next = fo->prev = NULL;
		fcur = folder;
	    } else {
		for (t = folder; t->next; t = t->next);
		t->next = fo;
		fo->next = NULL;
		fo->prev = t;
	    }
	    fid++;

	    next_unread(fo);
	}
#ifdef	DEBUG
	else 
	    printf("Incompliant folder format. [%s]\n", idx[i]);
#endif
	
    }
    
    return i;
}

static int load_folder(char *s)
{
    return loader(s, 0);
}

static int load_postpone(char *s)
{
    return loader(s, 1);
}


/* if current message was removed, adjust it to next */

static int adjust_cur_message(Folder *fo)
{
    Mail *m;
    
    for (m = fo->mail; m; m = m->next)  {
	if ((m->id >= fo->cur) && !m->removed)  {
	    fo->cur = m->id;
	    return m->id;
	}
    }
    
    /* hit the massage bottom */

    fo->cur = 0;
    for (m = fo->mail; m; m = m->next)  {
	if (!m->removed)
	    fo->cur = m->id;
    }
    
    return fo->cur;
}

/* get folder pointer */

static Folder *folder_no(char *s)
{
    int    id;
    Folder *fo = NULL;

#ifdef	DEBUG
    if (s == NULL)  {
	printf("folder_no()\n");
	return NULL;
    }
#endif

    switch (*s)  {
    case '#':
	id = atoi(s+1);
	for (fo = folder; fo; fo = fo->next)  {
	    if (fo->fid == id)
		break;
	}
	break;

    case '.':
	if (isspace(s[1]) || !s[1]) 
	    fo = fcur;
	break;

    case '$':
	for (fo = folder; fo && fo->next; fo = fo->next);
	break;
	
    case '\\':
	s++;
	/* here ignores break statement. it's not bug. */

    default:
	for (fo = folder; fo; fo = fo->next)  {
	    if (!strcmp(fo->name, s))
		break;
	}
	break;
    }

    return fo;
}

/**********************************************************************/

static Folder *_fost, *_foed;	/* start folder, end folder */
static char *_folst;
static int  _msglst[PARMAX][2];
static int  _msgcur, _msglen;

static int lst_msg_number()
{
    int  rs;
    
    if (_msgcur == _msglen)
	return -1;

    if (_msglst[_msgcur][0] == _msglst[_msgcur][1])  {
	rs = _msglst[_msgcur][0];
	_msgcur ++;
    } else if (_msglst[_msgcur][0] < _msglst[_msgcur][1])  {
	rs = _msglst[_msgcur][0];
	_msglst[_msgcur][0]++;
    } else {
	rs = _msglst[_msgcur][0];
	_msglst[_msgcur][0]--;
    }
    return rs;
}

static Folder *lst_start_folder()
{
    return _fost;
}

static Folder *lst_end_folder()
{
    return _foed;
}

static char *lst_last_fld()
{
    return _folst;
}

static int lst_init(char *s, int lf)
{
    char *idx[PARMAX], *ndx[2];
    int  i, rs;
    
    _fost = _foed = NULL;
    _folst = NULL;
    _msgcur = _msglen = 0;

    if (!s)
	return 0;
    
    rs = ziptoken(s, idx, PARMAX, " ");
    
    if ((rs > 0) && lf)  {
	_folst = idx[rs - 1];
	_foed = folder_no(idx[rs - 1]);
	rs--;
    }
    
    i = 0;
    if (idx[i] && !isdigit(*idx[i])) {
	_fost = folder_no(idx[i]);
	i++;
    }

    for ( ; i < rs; i++)  {
	
	ziptoken(idx[i], ndx, 2, "-");
	_msglst[_msglen][0] = atoi(ndx[0]);
	if (ndx[1])
	    _msglst[_msglen][1] = atoi(ndx[1]);
	else
	    _msglst[_msglen][1] = _msglst[_msglen][0];
	
	_msglen++;
    }
    return _msglen;
}

#ifdef	DEBUG

static int lst_test()
{
    char buf[SVRBUF];
    int  rs;

    while (fgets(buf, SVRBUF, stdin)) {
	lst_init(buf,0);
	while ((rs = lst_msg_number()) >= 0)
	    printf("%d ", rs);
	printf("\n");
    }
    return 0;
}

#endif


/**********************************************************************/
	
static Folder *fld_number(char *s)
{
    int  id;
    Folder *fo = NULL;

    if (isdigit(*s))  {
	id = atoi(s);
	for (fo = folder; fo; fo = fo->next)  {
	    if (fo->fid == id)
		break;
	}
    } else 
	fo = folder_no(s);
    
    return fo;
}
	
/* -f <start> <end> */

static int list_folder(char *s)
{
    Folder *begin, *end;
    char   buf[SVRBUF], flag[6], *idx[2];
    int    len;
    
    begin = end = NULL;
    ziptoken(s, idx, 2, " ");
    if (idx[0])
	begin = fld_number(idx[0]);
    
    if (begin == NULL)
	begin = folder;
    else
	fcur = begin;	/* re-posite current folder */
    
    if (idx[1])  
	end = fld_number(idx[1]);
    
    if (begin == NULL)  
	printf("No folder loaded.\n");
    else {
	len = folder_num();
	printf("%d %s ready.\n", len, len < 2 ? "folder is" : "folders are");
    }

    for ( ; begin; begin = begin->next)  {
	
	if (fcur->fid == begin->fid)  
	    flag[0] = '>';
	else  
	    flag[0] = ' ';
	
	if (begin->fchg)  
	    flag[1] = '*';
	else
	    flag[1] = ' ';
	
	if (begin->pp)
	    flag[2] = flag[3] = 'p';
	else
	    flag[2] = flag[3] = ' ';
	
	flag[4] = 0;
	
	len = sprintf(buf, "%s #%-3d%4d messages%4d unread%4d removed  %s", 
		flag, begin->fid, begin->num, begin->unrd, 
		begin->rm, begin->name);
	
	if (len >= cols)  {
	    buf[cols - 2] = '-';
	    buf[cols - 1] = '\0';
	}
	puts(buf);

	if (end && (begin->fid == end->fid))
	    break;
    }
    return 0;
}

/****************** for list message in the folders ********************/

/*  MM-DD HH:MM, WDS YYYY
       or
    YYYY-MM-DD   HH:MM
*/
/*
static char *date_form(char *buf, time_t *t)
{
    struct tm  *td, *mt;
    char   *wds[] = { "Sunday   ", "Monday   ", "Tuesday  ", 
	    "Wednesday", "Thursday ", "Friday   ", "Saturday "  };

    mt = localtime(t);
    
    if (today->tm_year == mt->tm_year)  {
	sprintf(buf, "%02d-%02d %02d:%02d, %s %4d", 
		mt->tm_mon+1, mt->tm_mday, mt->tm_hour, mt->tm_min, 
		wds[mt->tm_wday], mt->tm_year + 1900);
    } else {
	sprintf(buf, "%4d-%02d-%02d   %02d:%02d", 
		mt->tm_year + 1900, mt->tm_mon+1,
		mt->tm_mday, mt->tm_hour, mt->tm_min);
    }
    return buf;
}
*/

static char *date_form(time_t *t)
{
    static char buf[32];
    struct tm  *mt;

    mt = localtime(t);
    
    if (today->tm_year == mt->tm_year)  {
	sprintf(buf, "%02d-%02d %02d:%02d", 
		mt->tm_mon+1, mt->tm_mday, mt->tm_hour, mt->tm_min);
    } else {
	sprintf(buf, "%4d-%02d-%02d ", 
		mt->tm_year + 1900, mt->tm_mon+1, mt->tm_mday);
    }
    return buf;
}

static char *size_form(unsigned int sz)
{
    static char buf[16];

    if (sz < 10000)
	sprintf(buf, "%u", sz);
    else if (sz < 1048576)
	sprintf(buf, "%uK", sz >> 10);
    else 
	sprintf(buf, "%uM", sz >> 20);
    
    return buf;
}

static char *id_form(Folder *fo, int id)
{
    static char buf[16];
    int    max;

    max = fo->num + fo->rm;

    if (max < 100) 
	sprintf(buf, "%2d", id);
    else if (max < 1000)  
	sprintf(buf, "%3d", id);
    else if (max < 10000) 
	sprintf(buf, "%4d", id);
    else 
	sprintf(buf, "%5d", id);  /* rarely */
    
    return buf;
}

static char *addr_form(char *addr, int lmt)
{
    static char buf[128];	/* should be enough */
    int    max;

    if (addr == NULL)  {
	max = sizeof(buf) - 1;
	memset(buf, ' ', max);
	buf[max] = '\0';
    } else {
	StrnCpy(buf, addr, sizeof(buf));
	max = strlen(buf);
    }

    if (max > lmt)  
	buf[lmt - 1] = '-';
    else 
	for ( ; max < lmt; buf[max++] = ' ');
    
    buf[lmt] = '\0';
    return buf;
}

	
/* -h <folder> <start> <end> */

static int list_message(char *s)
{
    Folder *fo;
    Mail   *m;
    int    begin, tops;
    char   flag[4], buf[SVRBUF], *from;

    lst_init(s,0);
    
    if ((fo = lst_start_folder()) == NULL) 
	fo = fcur;
    
    if (fo == NULL)  {
	printf("No specified folder found.\n");
	return -1;
    }
    
    if ((begin = lst_msg_number()) <= 0) {
	begin = fo->lst;
	if (begin <= 0)  
	    begin = 1;
    } else {
	if (begin > msg_last_actived(fo))  {
	    printf("[%s]: No applicable messages.\n", fo->name);
	    return -1;
	}   
	fo->lst = begin;
	fo->cur = begin;
    }
    
    tops = lst_msg_number() - begin + 1;
    if (tops <= 0)  
	tops = lines - 2;
    
    /* now begin displaying message headline */
    
    printf("[%s]: %d messages %s, %d unread %d removed\n", fo->name, 
	fo->num, fo->pp ? "postponed" : "actived", fo->unrd, fo->rm);
    
    /* move to start pointe */

    for (m = fo->mail; m && (m->id < begin); m = m->next);
    
    while (m && tops)  {
	
	if (m->removed && hidrm)  {
	    m = m->next;
	    continue;	/* move to next message but no decreasing */
	}

	if (m->id == fo->cur)
	    flag[0] = '>';  
	else
	    flag[0] = ' ';
	
//	if (m->oldmail || m->changed)  flag[1] = ' ';
//	else  flag[1] = 'U';
//	
	if (m->oldmail)
	    flag[1] = ' ';
	else
	    flag[1] = 'U';
	
	if (m->changed)
	    flag[1] = '*';
	if (m->replied)
	    flag[1] = 'R';
	if (m->removed)
	    flag[1] = 'D';
	if (m->piped)
	    flag[1] = '|';
	/* flag[2] = ' ';     not used yet */
	flag[2] = flag[3] = '\0';

	if (fo->pp && m->to)
	    from = m->to;
	else
	    from = m->from;
	
	date_form(&m->rtime);
	
	snprintf(buf, SVRBUF - 1, "%s %s %s %4s %s %s",  
		flag, id_form(fo, m->id), addr_form(from, cols / 4), 
		size_form(m->msg->msize), date_form(&m->rtime), 
		m->subject ? m->subject : "");
	
	buf[SVRBUF - 1] = '\0';
	
	if (cols < strlen(buf))
	    buf[cols - 1] = '-';
	buf[cols] = '\0';

	puts(buf);

	tops --;
	m = m->next;
	
    }	/* end while */
    
    adjust_cur_message(fo);
    
    return 0;
}


static int close_folders(char *s, int sv)
{
    int    i, rs, cid;
    char   *idx[PARMAX];
    Folder *fo;

    rs = 0;
    if (s && *s)
	rs = ziptoken(s, idx, PARMAX, " ");

    if (rs == 0)  {	/* quit all folders */
    
	fo = folder;
	while (fo)  {
	    if (fo->fchg && sv)
		msg_combine(fo);
	    msg_clean(fo);
	    fo = fo->next;
	    if (fo)
		free(fo->prev);
	}
	folder = fcur = NULL;
	fid = 1;
	
    } else {
    
	cid = 0;
	if (fcur)
	    cid = fcur->fid;

	/* close specified folder */
	
	for (i = 0; i < rs; i++)  {
	    if ((fo = folder_no(idx[i])) == NULL)
		continue;
	    if (fo->fchg && sv)
		msg_combine(fo);
	    msg_clean(fo);
	    
	    if (fo == folder)  {
		folder = fo->next;
		if (fo->next)
		    fo->next->prev = NULL;
	    } else {
	    	fo->prev->next = fo->next;
	    	if (fo->next)
		    fo->next->prev = fo->prev;
	    }
	    free(fo);
	}

	/* adjust current folder */
	
	for (fo = folder; fo; fo = fo->next)  
	    if (fo->fid >= cid)
		break;
	if ((fcur = fo) == NULL)
	    fcur = folder;

    }
    return 0;
}

	
/* -q <folder list> */

static int quit(char *s)
{
    return close_folders(s, 1);
}
	
 /* -x <folder list> */

static int giveup(char *s)
{
    return close_folders(s, 0);
}

/******************* routines for dumping message *************/


/* dump file in the "pdmsg" */

static void mailz_dump(char *pager)
{
    FILE  *fp;
    char  buf[SVRBUF];


    if (pdmsg == NULL)
	return;
    
    if (pager == NULL)  {
	if ((fp = fopen(pdmsg, "r")) == NULL)
	    return;
	
	while (fgets(buf, SVRBUF, fp))  
	    printf("%s", buf);
	
	fclose(fp);
	return;
    }

    if (!strcmp(pager, "/bin/more") || !strcmp(pager, "/usr/bin/more"))
	sprintf(buf, "%s %s", pager, pdmsg);
    else 
	sprintf(buf, "cat %s | %s", pdmsg, pager);

    system(buf);
}
	
static int mime_decode(Mail *m)
{
    FILE  *fin, *fout;
    char  buf[SVRBUF];

    /* it's a dirty solution: decode message first to a file xxxxx.000,
     * then translate it back to file xxxxx, delete xxxxx.000, finally
     * call the pager to display file xxxxx  */
    
    if ((fin = fopen(m->msg->name, "r")) == NULL)  {
	dbgPrintf("Cannot open message for dumping. [%s]\n", m->msg->name);
	return 1;
    }

    head_clean(phead);
    head_of_file(phead, fin, buf, SVRBUF);

    mime_body(m, fin, pdmsg, 0);	
    
    fclose(fin);
    
    if ((fin = fopen(pdmsg, "r")) == NULL)
	return 1;
    unlink(pdmsg);
    if ((fout = fopen(pdmsg, "w")) == NULL)  {
	fclose(fin);
	return 2;
    }
	    
    fprintf(fout, "Message %d:\n", m->id);
    
    if ((mime == 2) && headmsk)
	head_mask(phead, headmsk, 0);

    /* In fact, it hacks mhead.c file, not good thing */
    /* head_dump(phead, fout, 1); */
    {
	HLin  *hl;
	struct head_chain *hc;

	if (phead->fmsp)
	    fputs(phead->fmsp, fout);
	for (hl = phead->hlin; hl; hl = hl->next)  {
	    if (hl->rm)
		continue;
	    if (hl->headin)  {
		StrnCpy(buf, hl->headin, SVRBUF);
		decode_head(buf, 2);
		fprintf(fout, "%s\n", buf);
	    }
	    for (hc = hl->hc; hc; hc = hc->next)  {
		StrnCpy(buf, hc->str, SVRBUF);
		decode_head(buf, 2);
		fprintf(fout, "%s\n", buf);
	    }
	}
    }	
    fputs("\n", fout);
    
    dup_stream(fin, fout);

    fclose(fin);
    fclose(fout);
    return 0;
}


static int dump_mail(Folder *fo, int mid)
{
    Mail  *m;
    
    if (fo == NULL)  {
	printf("Out of folders.\n");
	return 1;
    }

    for (m = fo->mail; m; m = m->next)  {
	if (m->id == mid)
	    break;
    }

    if (!m || m->removed)  {
	printf("Out of messages. [%d]\n", mid);
	return 1;
    }

    if (!m->oldmail)  {
	fo->unrd--;
	m->changed = 1;
	m->oldmail = 1;
    }
    
    if ((mime == 0) || (phead == NULL) || (pdmsg == NULL)) 
	msg_dump(m, pager);
    else if (mime_decode(m) == 0)
	mailz_dump(pager);
    else
	msg_dump(m, pager);
    
    if (m->changed)
	fo->fchg = 1;
    
    return 0;
}   

/* -p <folder> <msg1> <msg2> ... */

static int print(char *s)
{
    Folder  *fo;
    int     rs, cnt;
    

    lst_init(s,0);

    if ((fo = lst_start_folder()) == NULL)
	fo = fcur;
    if (fo == NULL)  {
	printf("No specified folder found.\n");
	return -1;
    }

    cnt = 0;
    if ((rs = lst_msg_number()) < 0)
	rs = fo->cur;
    do {
	if (!dump_mail(fo, rs))  
	    cnt++;
    } while ((rs = lst_msg_number()) >= 0);

    printf("Folder [%s]: %d messages are dumped.\n", fo->name, cnt);
    return cnt;
}


/* -a <folder> <msg1> <msg2> ... */

static int add_addr(char *s)
{
    Folder  *fo;
    Mail    *m;
    Memo    *memo;
    int     rs, cnt;
    
    if (addrmemo == NULL)  {
	dbgPrintf("No address memo specified.\n");
	return 0;
    }
    
    lst_init(s,0);

    if ((fo = lst_start_folder()) == NULL)
	fo = fcur;
    if (fo == NULL)  {
	printf("No specified folder found.\n");
	return -1;
    }

    if ((memo = memo_open(addrmemo)) == NULL)  
	return 0;
    
    cnt = 0;
    if ((rs = lst_msg_number()) < 0)
	rs = fo->cur;
    do {
	if ((m = msg_search_id(fo, rs)) == NULL)  
	    printf("Folder [%s] message %d: not found.\n", fo->name, rs);
	else
	    cnt += memo_add_addr(memo, m->from);
    } while ((rs = lst_msg_number()) >= 0);
	    
    memo_close(memo);
    printf("Folder [%s]: %d addresses are recorded.\n", fo->name, cnt);
    return cnt;
}

/* -d <folder> <msg1> <msg2> ... */

static int delete(char *s)
{
    Folder  *fo;
    Mail    *m;
    int     rs, cnt;
    
    lst_init(s,0);

    if ((fo = lst_start_folder()) == NULL)
	fo = fcur;
    if (fo == NULL)  {
	printf("No specified folder found.\n");
	return -1;
    }
    
    cnt = 0;
    if ((rs = lst_msg_number()) < 0)
	rs = fo->cur;
    do {
	if ((m = msg_search_id(fo, rs)) == NULL)
	    printf("Folder [%s] message %d: not found.\n", fo->name, rs);
	else if (m->removed)  
	    printf("Folder [%s] message %d: removed.\n", fo->name, rs);
	else {
	    m->removed = 1;
	    if (!m->oldmail)
		fo->unrd--;
	    cnt++;
	    fo->rm++;
	    fo->num--;
	}	
    } while ((rs = lst_msg_number()) >= 0);

    adjust_cur_message(fo);
    fo->fchg = 1;
    printf("Folder [%s]: %d messages are removed\n", fo->name, cnt);
    return cnt;
}

/*  move message to another folder */
/*  -m <mlist> folder */

static int move(char *s)
{
    Folder *fo;
    char   *p;
    int    rs, cnt;
    Mail   *m;

    if (fcur == NULL)  {	/* source folder */
	printf("Out of folder.\n");
	return -1;
    }
    
    lst_init(s,1);

    if ((p = lst_last_fld()) == NULL)  {	/* target folder */
	printf("No target folder specified.\n");
	return -2;
    }
    fo = lst_end_folder();
    
    cnt = 0;
    if ((rs = lst_msg_number()) < 0)
	rs = fcur->cur;
    do {
	if ((m = msg_pickout(fcur, rs)) == NULL)  {
	    printf("Message #%d not found.\n", rs);
	    continue;
	}

	if (fo) {
	    msg_insert(fo, m);
	    cnt++;
	} else  {
	    if (msg_appendto(m, p) == PM_MSG_OK)  
		cnt++;
	    msg_free_mailobj(m);
	    free(m);    
	}

    } while ((rs = lst_msg_number()) >= 0);

    adjust_cur_message(fcur);
    printf("Folder [%s]: %d messages were moved\n", fcur->name, cnt);
    return cnt;
}

/* -s <mlist> folder */

static int save(char *s)
{
    Folder *fo;
    char   *p;
    int    rs, cnt;
    Mail   *m;

    if (fcur == NULL)  {	/* source folder */
	printf("Out of folder.\n");
	return -1;
    }
    
    lst_init(s,1);

    if ((p = lst_last_fld()) == NULL)  {	/* target folder */
	printf("No target folder specified.\n");
	return -2;
    }
    fo = lst_end_folder();
    
    cnt = 0;
    if ((rs = lst_msg_number()) < 0)
	rs = fcur->cur;
    do {
	if ((m = msg_search_id(fcur, rs)) == NULL)  {
	    printf("Message #%d not found.\n", rs);
	    continue;
	}
	    
	if (fo)  {
	    if (msg_expand(fo, gdir, m) != NULL) 
		cnt++;
	} else {
	    if (msg_appendto(m, p) == PM_MSG_OK)  
		cnt++;
	}
    } while ((rs = lst_msg_number()) >= 0);
	
    printf("Folder [%s]: %d messages were saved\n", fcur->name, cnt);
    return cnt;
}

/* check out the postpone folder, -1 means no postpone */

static Folder *find_postpone()
{
    Folder *fo;
    char   buf[SVRBUF];
    
    for (fo = folder; fo; fo = fo->next)  
	if (fo->pp)  break;
    
    if (fo == NULL)  {
	StrnCpy(buf, postpone, SVRBUF);
	load_postpone(buf);
	for (fo = folder; fo; fo = fo->next)  
	    if (fo->pp)  break;
    }
    
    return fo;
}
	

/* -r <mlist> */

static int reply(char *s)
{
    Folder  *fo;
    Mail    *m, *sour;
    int     rs, cnt;
    char    buf[SVRBUF];
    FILE    *fin, *fout;
    time_t  dc;

    if (fcur == NULL)  {
	printf("Out of folder.\n");
	return -1;
    }

    if ((fo = find_postpone()) == NULL)  {
	printf("No postpone folder defined.\n");
	return -2;
    }
   
    lst_init(s,0);
    cnt = 0;
    if ((rs = lst_msg_number()) < 0)
	rs = fcur->cur;
    do {
	sour = msg_search_id(fcur, rs);
	if (!sour || sour->removed)  {
	    printf("Message #%d not found.\n", rs);
	    continue;
	}
    	if ((m = msg_expand(fo, gdir, NULL)) == NULL)  
	    continue;
     
	/*m->from      = dup_str(replier);	 my address */
	if (sour->to) 
	    m->from  = dup_str(sour->to);
	else
	    m->from  = dup_str(replier);
	    
	m->reply     = dup_str(replier);
	m->to	     = dup_str(sour->reply);
	m->cc 	     = dup_str(sour->cc);
	m->subject   = dup_str(sour->subject);
	m->rtime     = time(NULL);
	m->msg->len  = sour->msg->len;
	m->msg->line = sour->msg->line;

	/* copy original mail to temporary folder */
	
	if ((fin = fopen(sour->msg->name, "r")) == NULL)  
	    continue;
	fseek(fin, sour->msg->boff, SEEK_SET);
	mime_body(sour, fin, pdmsg, 1);
	fclose(fin);
	if ((fin = fopen(pdmsg, "r")) == NULL)  
	    continue;
	    
	if ((fout = fopen(m->msg->name, "w")) == NULL)  {
	    fclose(fin);
	    continue;
	}

	fprintf(fout, "From %s %s", email_pure(replier), ctime(&m->rtime));
	/*fprintf(fout, "From: %s\n", replier);*/
	fprintf(fout, "From: %s\n", m->from);
	fprintf(fout, "To: %s\n", m->to);
	if (sour->cc)  
	    fprintf(fout, "Cc: %s\n", sour->cc);
	fprintf(fout, "Subject: Re: %s\n\n", m->subject);
	m->msg->boff = ftell(fout);
		
	fprintf(fout, "%s", quota);
	while (fgets(buf, SVRBUF, fin))  
	    fprintf(fout, "> %s", buf);

	if (ensig && signature)  
	    fprintf(fout, "%s", signature);
	
	fclose(fin);
	fclose(fout);

	m->msg->mtime = filedate(m->msg->name, &m->msg->msize);
	if (m->msg->mtime < 0)  
	    m->removed = 1;
	else {
	    sprintf(buf, "%s %s", editor, m->msg->name);
	    system(buf);
	    
	    dc = filedate(m->msg->name, &m->msg->msize);
	    if (dc > m->msg->mtime)  {
		sour->replied = 1, cnt++;
		m->msg->mtime = dc;
	    } else  
		m->removed = 1;
	}
	
	m->changed = 1;
	if (m->removed)  {
	    fo->rm++;
	    fo->num--;
	}
	if (!m->oldmail)  {
	    m->oldmail = 1;
	    fo->unrd--;
	}
	if (!sour->oldmail)  {
	    sour->oldmail = 1;
	    sour->changed = 1;
	    fcur->unrd--;
	    fcur->fchg = 1;
	}
    } while ((rs = lst_msg_number()) >= 0);
	    
    printf("Folder [%s]: %d messages were replyed\n", fcur->name, cnt);
    return cnt;
}

/* -fo <mlist> */

static int follow(char *s)
{
    Folder  *fo;
    Mail    *m, *sour;
    int     rs, cnt;
    char    buf[SVRBUF];
    FILE    *fin, *fout;
    time_t  dc;

    if (fcur == NULL)  {
	printf("Out of folder.\n");
	return -1;
    }

    if ((fo = find_postpone()) == NULL)  {
	printf("No postpone folder defined.\n");
	return -2;
    }

    lst_init(s,0);
    cnt = 0;
    if ((rs = lst_msg_number()) < 0)
	rs = fcur->cur;
    do {
	sour = msg_search_id(fcur, rs);
	if (!sour || sour->removed)  {
	    printf("Message #%d not found.\n", rs);
	    continue;
	}
    	if ((m = msg_expand(fo, gdir, NULL)) == NULL)  
	    continue;
     
	m->from      = dup_str(replier);	 
	m->reply     = dup_str(sour->reply);
	m->to        = dup_str(sour->to);
	m->cc        = dup_str(sour->cc);
	m->subject   = dup_str(sour->subject);
	m->rtime     = time(NULL);
	m->msg->len  = sour->msg->len;
	m->msg->line = sour->msg->line;

	/* copy original mail to temporary folder */
	
	if ((fin = fopen(sour->msg->name, "r")) == NULL)  
	    continue;
	fseek(fin, sour->msg->boff, SEEK_SET);
	mime_body(sour, fin, pdmsg, 1);
	fclose(fin);
	if ((fin = fopen(pdmsg, "r")) == NULL)  
	    continue;
	    
	if ((fout = fopen(m->msg->name, "w")) == NULL)  {
	    fclose(fin);
	    continue;
	}

	fprintf(fout, "From %s %s", email_pure(replier), ctime(&m->rtime));
	/*fprintf(fout, "From: %s\n", replier);*/
	fprintf(fout, "From: %s\n", m->from);
	fprintf(fout, "To: %s\n", m->to);
	fprintf(fout, "Reply-to: %s\n", m->reply);
	if (sour->cc)  
	    fprintf(fout, "Cc: %s\n", sour->cc);
	fprintf(fout, "Subject: FO: %s\n\n", m->subject);
	m->msg->boff = ftell(fout);
		
	fprintf(fout, "%s", quota);
	while (fgets(buf, SVRBUF, fin))  
	    fprintf(fout, "> %s", buf);

	if (ensig && signature)  
	    fprintf(fout, "%s", signature);
	
	fclose(fin);
	fclose(fout);

	m->msg->mtime = filedate(m->msg->name, &m->msg->msize);
	if (m->msg->mtime < 0)  
	    m->removed = 1;
	else {
	    sprintf(buf, "%s %s", editor, m->msg->name);
	    system(buf);
	    
	    dc = filedate(m->msg->name, &m->msg->msize);
	    if (dc > m->msg->mtime)  {
		sour->replied = 1, cnt++;
		m->msg->mtime = dc;
	    } else  
		m->removed = 1;
	}
	
	m->changed = 1;
	if (m->removed)  {
	    fo->rm++;
	    fo->num--;
	}
	if (!m->oldmail)  {
	    m->oldmail = 1;
	    fo->unrd--;
	}
	if (!sour->oldmail)  {
	    sour->oldmail = 1;
	    sour->changed = 1;
	    fcur->unrd--;
	    fcur->fchg = 1;
	}
    } while ((rs = lst_msg_number()) >= 0);
    
    printf("Folder [%s]: %d messages were followed\n", fcur->name, cnt);
    return cnt;
}

/* -c <address_list> */

static int composer(char *s)
{
    Folder *fo;
    Mail   *m;
    Memo   *memo;
    FILE   *fp;
    char   buf[SVRBUF];
    time_t dc;

    if ((fo = find_postpone()) == NULL)  {
	printf("No postpone folder defined.\n");
        return -2;
    }

    if ((m = msg_expand(fo, gdir, NULL)) == NULL)  return -3;

    m->from  = dup_str(replier); 
    m->reply = dup_str(replier);
    
    if (!s || isblankline(s))  {
	memo = memo_open(addrmemo);
	s = memo_choose(memo, cols, lines);
	if (s == (char*) memo_choose)  {
	    m->removed = 1;
	    return 0;
	}
	m->to = dup_str(s);
	memo_close(memo);
    }

    fp = fopen(m->msg->name, "w");
    if (fp == NULL)  {
	dbgPrintf("Cannot compose message. [%s]\n", m->msg->name);
	m->removed = 1;
	return -1;
    }

    fprintf(fp, "From %s %s", email_pure(replier), ctime(&m->rtime));
    fprintf(fp, "From: %s\n", replier);
    fprintf(fp, "To: %s\n", m->to ? m->to : " ");
    fprintf(fp, "Subject:\n\n");
    m->msg->boff = ftell(fp);

    if (ensig && signature)  fprintf(fp, "\n%s\n", signature);
    fclose(fp);
    
    m->msg->mtime = filedate(m->msg->name, &m->msg->msize);
    if (m->msg->mtime < 0)  m->removed = 1;
    else {
	sprintf(buf, "%s %s", editor, m->msg->name);
	system(buf);
	dc = filedate(m->msg->name, &m->msg->msize);
	if (dc > m->msg->mtime)  m->msg->mtime = dc;
	else  m->removed = 1;
    }
    m->changed = 1;
    if (m->removed)  {
	fo->rm++;
	fo->num--;
    }
    if (!m->oldmail)  {
	m->oldmail = 1;
	fo->unrd--;
    }
    return 0;
}
    

/* -e <folder> <mlist> */

static int edit(char *s)
{
    Folder *fo;
    Mail   *m;
    int    rs, cnt;

    lst_init(s,0);

    if ((fo = lst_start_folder()) == NULL)
	fo = fcur;
    if (fo == NULL)  {
	printf("No specified folder found.\n");
	return -1;
    }

    cnt = 0;
    if ((rs = lst_msg_number()) < 0)
	rs = fo->cur;
    do {
	m = msg_search_id(fo, rs);
	if (!m || m->removed)  {
	    printf("Message #%d not found.\n", rs);
	    continue;
	}

	if (!m->oldmail)  {
	    fo->unrd--;
	    m->changed = 1;
	    m->oldmail = 1;
	}
	msg_edit(m, editor);
	if (m->changed) {
	    fo->fchg = 1;
	    cnt++;
	}
    } while ((rs = lst_msg_number()) >= 0);

    printf("Folder [%s]: %d messages were edited\n", fcur->name, cnt);
    return 0;
}


/* -u <folder> <mlist> */

static int undo(char *s)
{
    Folder *fo;
    Mail   *m;
    int    k, rs, cnt;

    lst_init(s,0);

    if ((fo = lst_start_folder()) == NULL)
	fo = fcur;
    if (fo == NULL)  {
	printf("No specified folder found.\n");
	return -1;
    }

    cnt = 0;
    if ((rs = lst_msg_number()) < 0) {

	/* No messages specified, so undo entire folder */
	for (m = fo->mail; m; m = m->next)  {
	    k = 0;
	    if (m->removed)  {
		m->removed = 0;
		fo->num++;
		fo->rm--;
		if (!m->oldmail)
		    fo->unrd++;
		k = 1;
	    }
	    if (m->piped) {
		msg_unpipe(m);
		k = 1;
	    }
	    cnt += k;
	}
	
    } else {
	do  {
	    if ((m = msg_search_id(fo, rs)) == NULL)  
		continue;
	    
	    k = 0;
	    if (m->removed)  {
		m->removed = 0;
		fo->rm--;
		fo->num++;
		if (!m->oldmail)  
		    fo->unrd++;
		k = 1;
	    }
	    if (m->piped) {
		msg_unpipe(m);
		k = 1;
	    }
	    cnt += k;
	} while ((rs = lst_msg_number()) >= 0);
    }
    
    adjust_cur_message(fo);
    printf("Folder [%s]: undo %d messages.\n", fo->name, cnt);
    return cnt;
}


/* -| command */

static int pipeto(char *s)
{
    Mail   *m;

    if (fcur == NULL)  {
	printf("Out of folder.\n");
	return -1;
    }

    if ((m = msg_search_id(fcur, fcur->cur)) == NULL)  {
	printf("No current message specified.\n");
	return -2;
    }
    
    msg_pipeto(m, s, 0, NULL);
    if (m->changed)  
	fcur->fchg = 1;
    return 0;
}

/* -! <command> */
	    
static int shell(char *s)
{
    return system(s);
}
 

static char *init_signature(char *s)
{
    FILE  *fsig;
    char  buf[SVRBUF], *sig;
    int	  len;
    
    if ((fsig = fopen(buf, "r")) == NULL)  {
	logPrintf(LOG_WARN, 
		"mailz: can not open signature file. [%s]\n", buf);
	return NULL;
    }
    
    fseek(fsig, 0, SEEK_END);
    if ((len = ftell(fsig)) > MAX_SIG)  {
	logPrintf(LOG_WARN, 
		"mailz: Too large is the signature file. [%d]\n", len);
	fclose(fsig);
	return NULL;
    }
    
    /*fprintf(stderr, "%s [%d]\n", buf, len);*/
    
    rewind(fsig);
    if ((sig = (char*) malloc(len + 4)) == NULL)  {
	fclose(fsig);
	logPrintf(LOG_WARN,
		"mailz: Out of signature memory. [%d]\n", len);
	return NULL;
    }

    fread(sig, 1, len, fsig);
    
    fclose(fsig);
    return sig;
}	       
    

static char *hlp1 = "Usage: mailz [-p postpone] [folder ...]\n\n";
static char *hlp2 = "\
a     <folder> <mlist>  add up sender addresses to the address book\n\
c     <address_list>    compose a message\n\
d     <folder> <mlist>  delete messages in the specified folder\n\
e     <folder> <mlist>  edit these messages in the specified folder\n\
f     <folder_range>    list loaded folders, and activate first folder\n\
fo    <mlist>           follow this message\n\
h     <folder> <range>  list active message heads in the specified folder\n\
l     folder_list       load mail folders\n\
lpp   postpone_list     load postpone folders\n\
m     <mlist> folder    move messages to the specified folder\n\
p     <folder> <mlist>  print these messages on the screen\n\
q     <folder_list>     quit or close specified folders and save changes\n\
r     <mlist>           reply to message originor and all recipients\n\
s     <mlist> folder    save messages to specified folder\n\
set   <key=option>      change options, see 'help set'\n\
u     <folder> <mlist>  undo removed or modified messages\n\
x     <folder_list>     quit or close specified folders without saving\n\
!     <command>         shell escape\n\
|     command           require external pipe\n\
?     <version | set>   help information\n\n";

static char *hlp3 = "\
A <range> consists of start number and end number.\n\
The \"mlist\" is the list of message number.\n\
The \"folder\" can be folder number started with '#'. Without the '#' it would\n\
  be looked as folder name even though a number. e.g. '#12' is the 12th folder\n\
  and '12' is a message folder named '12'. The char '.' means current folder;\n\
  the '$' means last folder.\n\
";

static char *set = "\
set [reply=string]          set return mail address\n\
set [pager=string]          set default pager for message browing\n\
set [editor=string]         set default editor\n\
set [postpone=string]       set default postpone folder\n\
set [memo=string]           set address memo\n\
set [lines=single|double]   set message head displaying\n\
set [hider=yes|no]          hide removed message\n\
set [signature=on|off]      enable or disable external signature file\n\
set [ps1=char]              set command line prompt character\n\
set [mime=off|full|simple]  set MIME decode action. 'full' keeps full message\n\
                            head; 'simple' removes less sense head.\n\
";
    
static char *ver = 
"mailz %s, copyright by Xuming <xuming@iname.com>, by GPL\n";

static int help(char *s)
{
    if (s == NULL)  
	printf(hlp1);
    else if (!strcmp(s, "version"))  
	printf(ver, VERSION);
    else if (!strcmp(s, "set"))  
	printf(set);
    else if (!strcmp(s, "cmd")) {
	printf(hlp1);
	printf(hlp2);
	printf(hlp3);
    } else 
	printf(hlp2);
    return 0;
}
     	

static int setinfo(char *s)
{
    char  *idx[2];
    int   rs;
    char  *mmsg[3] = { "off", "full-head", "simple-head" };

    rs = ziptoken(s, idx, 2, "= ");
    if (rs < 2) {
	/* display current option */

	printf("address memorandum:   %s\n", addrmemo);
	printf("external editor:      %s\n", editor);
	printf("external pager:       %s\n", pager);
	printf("MIME decoder:         %s\n", mmsg[mime]);
	printf("postpone folder:      %s\n", postpone);
	printf("prompt character:     \"%c\"\n", ps1);
	printf("removed message:      %s\n", hidrm ? "hide" : "mark");
	printf("reply address:        %s\n", replier);
	printf("signature:            %s\n", ensig ? "on" : "off");
	return 0;
    }

    if (!strcasecmp(idx[0], "reply"))  {
	if (replier)  
	    free(replier);
	replier = dup_str(idx[1]);
    } else if (!strcasecmp(idx[0], "pager"))  {
	if (pager)  
	    free(pager);
	pager = dup_str(idx[1]);
    } else if (!strcasecmp(idx[0], "editor"))  {
	if (editor)  
	    free(editor);
	editor = dup_str(idx[1]);
    } else if (!strcasecmp(idx[0], "postpone"))  {
	if (postpone)  
	    free(postpone);
	postpone = dup_str(idx[1]);
    } else if (!strcasecmp(idx[0], "memo"))  {
	if (addrmemo)  
	    free(addrmemo);
	addrmemo = dup_str(idx[1]);
    } else if (!strcasecmp(idx[0], "hider"))  {
	if (!strcmp(idx[1], "1") || !strcasecmp(idx[1], "yes")) 
	    hidrm = 1;
	else if (!strcmp(idx[1], "0") || !strcasecmp(idx[1], "no"))
	    hidrm = 0;
    } else if (!strcmp(idx[0], "signature"))  {
	if (!strcmp(idx[1], "0") || !strcasecmp(idx[1], "off"))
	    ensig = 0;
	else if (!strcmp(idx[1], "1") || !strcasecmp(idx[1], "on"))
	    ensig = 1;
    } else if (!strcasecmp(idx[0], "ps1"))  {
	ps1 = *idx[1];
    } else if (!strcasecmp(idx[0], "mime"))  {
	if (!strcmp(idx[1], "0") || !strcasecmp(idx[1], "off"))  
	    mime = 0;
	else if (!strcmp(idx[1], "1") || !strcasecmp(idx[1], "full"))  
	    mime = 1;
	else if (!strcmp(idx[1], "2") || !strcasecmp(idx[1], "simple"))  
	    mime = 2;
    } else 
	printf("Invalid action.\n");
    
    return 0;
}


struct  CMDLST  {
	
	char    *key;
	int     (*cmd)(char *s);
	
} cmdlst[] = {

	{ "a",     add_addr },
	{ "c",     composer },
	{ "lpp",   load_postpone },
	{ "l",     load_folder },
	{ "fo",    follow },
	{ "f",     list_folder },
	{ "help",  help },
	{ "h",     list_message },
	{ "x",     giveup },
	{ "p",     print },
	{ "q",     quit },
	{ "d",     delete },
	{ "m",     move },
	{ "set",   setinfo },
	{ "s",     save },
	{ "r",     reply },
	{ "e",     edit },
	{ "u",     undo },
	{ "|",     pipeto },
	{ "!",     shell },
	{ "?",     help },
	{ NULL,    NULL }
};

static void sig_quit(int sig)
{
    printf("signal %d encountered.\n", sig);
    giveup(NULL);
    free(pager);
    free(editor);
    free(postpone);
    free(replier);
    free(folder);
    free(signature);
    free(addrmemo);
    
    if (pdmsg)  {
	unlink(pdmsg);
	free(pdmsg);
    }
    head_destruct(phead);
    rmdir(gdir);
    exit(0);
}


static int  open_quota_log(char *s)
{
    char *idx[2];
    char buf[SVRBUF];
    int  rs, quota;

    quota = 0;
    rs = ziptoken(s, idx, 2, DELIMS);
    fullpath(idx[0], buf);
    if (idx[1])
        quota = atonum(idx[1]);
    
    return log_open(buf, quota);
}


static char *rcs[] = {
//	"mailzrc",
	"~/.mailzrc",
	"/etc/mailzrc",
//        "/usr/lib/mailzrc",
//        "/usr/local/etc/mailzrc",
//        "/usr/local/lib/mailzrc",
//        "pmailrc",
        "~/.pmailrc",
        "/etc/pmailrc",
//        "/usr/lib/pmailrc",
//        "/usr/local/etc/pmailrc",
//        "/usr/local/lib/pmailrc",
	NULL
};

static void init_option(char *pp)
{
    struct winsize ws;
    int    i;
    char   buf[SVRBUF], *p;
    time_t td;

	    
    for (i = 0; rcs[i]; i++)  {
	if (*rcs[i] == '~')  
	    sprintf(buf, "%s%s", get_home(), &rcs[i][1]);
	else  
	    strcpy(buf, rcs[i]);
	
	if (rc_open(buf) >= 0)  
	    break;
    }

    if (rcs[i])  {
	    
	if ((p = rc_check(RCKEY_PAGER)) != NULL)  {
	    fullpath(p, buf);
	    pager = dup_str(buf);
	}
	
	if ((p = rc_check(RCKEY_EDITOR)) != NULL)  {
	    fullpath(p, buf);
	    editor = dup_str(buf);
	}
	
	if ((p = rc_check(RCKEY_ADMEMO)) != NULL)  {
	    fullpath(p, buf);
	    addrmemo = dup_str(buf);
	}
	
	if ((p = rc_check(RCKEY_SIG)) != NULL)  {
	    fullpath(p, buf);
	    signature = init_signature(buf);
	}
	
	if ((p = rc_check(RCKEY_REPLY)) != NULL)
	    replier = dup_str(p);

	if (!folder && (p = rc_check(RCKEY_MBOX)) != NULL)  {
	    fullpath(p, buf);
	    load_folder(buf);
	}

	if (!postpone && (p = rc_check(RCKEY_OUTBOX)) != NULL)  {
	    fullpath(p, buf);
	    postpone = dup_str(buf);
	}

	if ((p = rc_check(RCKEY_DEBUG)) != NULL)  {
	    fullpath(p, buf);	
	    open_quota_log(buf);
	}

	if ((p = rc_check(RCKEY_HEADMASK)) != NULL)  {
	    headmsk = dup_str(buf);
	}
	
	rc_close();
    }
	
    if (!postpone)  {
	fullpath(OUTFOLDER, buf);
	postpone = dup_str(buf);
    }
    if (pp)  {
	fullpath(pp, buf);
	postpone = dup_str(buf);
	load_postpone(buf);
    }
    
    if (!pager)  {
	if ((p = getenv("PAGER")) == NULL)  
	    pager = dup_str(PAGER);
	else  
	    pager = dup_str(p);
    }
    if (!editor)  {
	if ((p = getenv("EDITOR")) == NULL)  
	    editor = dup_str(EDITOR);
	else 
	    editor = dup_str(p);
    }
    if (!replier)  {
	sprintf(buf, "%s@%s", get_user(), get_hostname());
	replier = dup_str(buf);
    }
    if (headmsk)  {
	headmsk = dup_str(HMDEF);
    }
    
    if (ioctl(1, TIOCGWINSZ, (char *) &ws) >= 0)  {
	cols = ws.ws_col;
	lines = ws.ws_row;
    }
    td = time(NULL);
    today = localtime(&td);
}

static char  *welcome = "\n\
Mailz version %s. Copyright by Xuming <xuming@iname.com>\n\
MAILZ is free software, covered by the GNU General Public License, and you are\n\
welcome to change it and/or distribute copies of it under certain conditions;\n\
check 'COPYING' for the conditions, which should come along with this program.\n\
There is absolutely no warranty for MAILZ.  See also 'COPYING' for details.\n\n\
";

int main(int argc, char **argv)
{
    Folder *fo;
    int   c, len, df, lr, uhlen;
    char  buf[SVRBUF], *usrhome, *p, *pp = NULL;
    char  *sopt = "hVp:";
    struct option lopt[] = {
	{ "help", 0, NULL, 'h' },
	{ "version", 0, NULL, 'V' },
	{ NULL, 0, NULL, 0 }
    };

    while ((c = getopt_long(argc, argv, sopt, lopt, NULL)) != EOF)  {
	switch (c) {
	case 'h':
	    help("cmd");
	    return 0;

	case 'V':
	    help("version");
	    return 0;

	case 'p':
	    pp = optarg;
	    break;
	    
	default:
	    printf("Unknown option [-%c]\n", c);
	    return 1;
	}
    }

    /*lst_test();
    return 0;*/

    signal(SIGHUP,  sig_quit);
    signal(SIGQUIT, sig_quit);
    signal(SIGABRT, sig_quit);
    signal(SIGINT,  sig_quit);
		    
    /* create temporary messages directory  */
    
    umask(0077);
    if (mktmpdir(gdir) == NULL)  {
	dbgPrintf("mailz: make temporary directory failed. [%s]\n", gdir);
	return -1;
    }

    /* create public Head object and dumping Mail object */

    phead = head_register(HDPAGE, HDIDX);
    len = strlen(gdir) + strlen(SHOWMSG) + 5;
    if ((pdmsg = malloc(len)) != NULL)  {
	sprintf(pdmsg, "%s/%s", gdir, SHOWMSG);
	/* printf("debug: %s\n", pdmsg); */
    }
    
    for ( ; optind < argc; optind++)  {
	load_folder(argv[optind]);
    }
    init_option(pp);
    
    
    printf(welcome, VERSION);
    folder_statis();
    if (folder)  {
	fcur = folder;
	list_message("");
    }

    lr = 0;	/* last input is not Enter */
    usrhome = get_home();
    uhlen = strlen(usrhome);
    
    while (1)  {
	
	
	if (fcur == NULL)  
	    sprintf(buf, "%c ", ps1);
	else if (!strncmp(usrhome, fcur->name, uhlen)) 
	    sprintf(buf, "~%s:%d%c ", fcur->name + uhlen, fcur->cur, ps1);
	else
	    sprintf(buf, "%s:%d%c ", fcur->name, fcur->cur, ps1);

	
#ifdef	HAVE_LIBREADLINE
	
	if ((p = readline(buf)) == NULL)  {
	    printf("\n");
	    continue;
	}
	StrnCpy(buf, p, SVRBUF);
	if (p[0])  
	    add_history(p);
	free(p);

#else
	printf("%s", buf);
	if (!fgets(buf, SVRBUF, stdin))  {
	    printf("\n");
	    continue;
	}
	buf[strlen(buf) - 1] = '\0';

#endif

	/* index the command in the command list */
	
	p = skip_space(buf);
	for (c = 0; cmdlst[c].key; c++)  {
	    len = strlen(cmdlst[c].key);
	    if (!strncmp(p, cmdlst[c].key, len))  
		break;
	}

	if (cmdlst[c].key)  {

	    lr = 0;
	    while (isspace(p[len]))  
		len++;
	    cmdlst[c].cmd(p + len);
	    
	    if (!strcmp(buf, "x"))  
		break;
	    if (!strcmp(buf, "q"))  
		break;
	    
	} else if (isdigitword(p))  {
		
	    df = atoi(p);
	    if (fcur && !dump_mail(fcur, df))  
		fcur->cur = lr = df;
	    
	} else if (*p == '\0')  {

	    if (fcur)  {
		if (lr)  {
		    fcur->cur++;
	    	    adjust_cur_message(fcur);
		}
		dump_mail(fcur, fcur->cur);
	    	/* fcur->cur++;*/
	    }
	    lr = 1;
		
	} else if ((fo = folder_no(p)) != NULL)  {

	    fcur = fo;
	    list_message("");
	    lr = 0;
		
	} else {
	    printf("Unknown command! [%s]\n", buf);
	    lr = 0;
	}
    }
    
    free(pager);
    free(editor);
    free(postpone);
    free(replier);
    free(folder);
    free(signature);
    free(addrmemo);
    free(headmsk);
    if (pdmsg)  {
	unlink(pdmsg);
	free(pdmsg);
    }
    head_destruct(phead);
    if (rmdir(gdir) < 0)  
	perror("rmdir");

    return 0;
}
    

