
/*
   pop3.c -- the routines for pop3 protocol.
	
   Copyright (C) 1998, 1999  Xuming <xuming@iname.com>
   
   This program is free software; you can redistribute it and/or 
   modify it under the terms of the GNU General Public License as 
   published by the Free Software Foundation; either version 2, or 
   (at your option) any later version.
	   
   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of 
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License, the file COPYING in this directory, for
   more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <termios.h>
#include <sys/time.h>
#include <sys/ioctl.h>

#include "defs.h"
#include "gmail.h"


static int check_gmail(GMAIL *gm, char *id, char *pass, int len);
static int pop3_connect(GMAIL *gm);
static int pop3_auth(GMAIL *gm, char *id, char *pass);
static int pop3_check(GMAIL *gm);
static int pop3_retr_one(GMAIL *gm, int id);
static int pop3_process_spam(GMAIL *gm, Spam *sc, Head *head, int sn);
static int pop3_spam_match(GMAIL *gm, Head *head, int sn);
static int pop3_retr(GMAIL *gm, Head *head);
static int pop3_list_basic(GMAIL *gm);
static void pop3_head_info(GMAIL *gm, Head *head, int sn);
static void pop3_head_digest(GMAIL *gm, Head *head);
static int pop3_list_head(GMAIL *gm, Head *head, int sn, int ctl);
static int pop3_list_top(GMAIL *gm, Head *head);
static int pop3_dele_one(GMAIL *gm, int id);
static int pop3_dele(GMAIL *gm);
static int pop3_quit(GMAIL *gm);
static int pop3_transact(int socket, char *cmd, int len);
	

int  pop3_main(GMAIL *gm, Head *head)
{
    int   rs;
    char  id[128], pass[128];
    
    if (check_gmail(gm, id, pass, 128) != OK)  
	return FAILURE;
    
    conPrintf("Trying POP3 server [%s], login as [%s] ...\n", 
		    gm->host, id);
    
    rs = open_socket(gm->host, POP3_PORT);
    if (rs < 0)  {
	dbgPrintf("Cannot connect to [%s].\n", gm->host);
	return rs;
    }
    gm->socket = rs;

    if ((rs = pop3_connect(gm)) != OK)  {
	close(gm->socket);
	gm->socket = 0;
	return rs;
    }
    
    if ((rs = pop3_auth(gm, id, pass)) != OK)  {
	memset(pass, 0, 128);
	close(gm->socket);
    	gm->socket = 0;
    	return rs;
    }
    memset(pass, 0, 128);
    
    if (gm->attr->delay)  
	sleep(gm->attr->delay);
    
    logPrintf(LOG_INFO, "pop3: login in %s as <%s>\n", gm->host, id);
    
    if ((rs = pop3_check(gm)) != OK)  {
	pop3_quit(gm);
    	close(gm->socket);
	gm->socket = 0;
    	return rs;
    }
    
    if (gm->msgs > 0)  {
	    
        conPrintf("%d messages ready, total %d bytes\n", gm->msgs, gm->volm);
	if (gm->act & GMAIL_LIST)  {
		
	    if (pop3_list_top(gm, head) == OK)
		logPrintf(LOG_ALL, "pop3: list %d messages in %s.\n", 
			gm->msgs, gm->host);
	    
	} else if (gm->act & GMAIL_FETCH)  {
		
	    if (pop3_retr(gm, head) == OK)  
		logPrintf(LOG_ALL, "pop3: download %d messages in %s.\n",
			gm->msgs, gm->host);
	    
	} else if (gm->act & GMAIL_DELE)  {
		
	    if (pop3_dele(gm) == OK)  
		logPrintf(LOG_ALL, "pop3: remove %d messages from server.\n",
			gm->msgs);
	}
    } else
	conPrintf("No message is ready\n");
	
    pop3_quit(gm);
    close(gm->socket);
    gm->socket = 0;
    return OK;
}

/********************************************************/
    
static int check_gmail(GMAIL *gm, char *id, char *pass, int len)
{    
    if (!gm->host || isblankline(gm->host))  
	return FAILURE;
    
    if (gm->lid)  
	StrnCpy(id, gm->lid, len);
    else  
	do {
            fprintf(stderr, "Login [%s] as: ", gm->host);
            fgets(id, len, stdin);
            chop(id);
        } while (isblankline(id));

    if (gm->pass)  {
	StrnCpy(pass, gm->pass, len);
	/*memset(gm->pass, 0, strlen(gm->pass));*/
	/* I cannot erase the gm->pass when gm->pass points to the
	 * attr->pass, that's a global password. */
    } else  {
	fprintf(stderr, "Enter key for [%s]: ", gm->host);
	strcpy(pass, getpass(""));
	fprintf(stderr, "\n");
    }
    return OK;
}
    	

static int pop3_connect(GMAIL *gm)
{
    char  buf[SVRBUF], *p;
    int	  rs;
    
    rs = sock_readall(gm->socket, buf, SVRBUF, timeout);
    if (rs != OK)  
	return rs;
    
    if (verbose)  
	conPrintf("<< %s", buf);
    
    if (*buf != '+')  {
    	dbgPrintf("gmail: what answer? %s", buf);
    	rs = pop3_quit(gm);
    	if (rs != OK)  
	    return rs;
    	return MISANSWER;
    }
   
    if ((p = strrchr(buf, '>')) == NULL)  
	return OK;
    *(++p) = '\0';
    if ((p = strchr(buf, '<')) == NULL)  
	return OK;
    gm->time_stamp = dup_str(p);
    
    /* conPrintf("Time stamp: %s\n", gm->time_stamp); */

    return OK;
}

static int pop3_auth(GMAIL *gm, char *id, char *pass)
{
    char  buf[SVRBUF], rb[16], *p;
    int   rs;
   
    if ((gm->protl == GP_APOP) && gm->time_stamp)  {
	rs = sprintf(buf, "APOP %s ", id);
	p = buf + rs;
	rs = sprintf(p, "%s%s", gm->time_stamp, pass);
	puts(buf);
	md5_buffer(p, rs, rb);
	md5_hex(rb, p);
	puts(buf);
	if ((rs = pop3_transact(gm->socket, buf, SVRBUF)) != OK)  {
	    dbgPrintf("APOP might not be supported by peer.\n");
	    /* when APOP failed, the server closes the connect */
	}
	return rs;
    }
    
    sprintf(buf, "USER %s", id);
    if ((rs = pop3_transact(gm->socket, buf, SVRBUF)) != OK)  {
	dbgPrintf("User %s is rejected.\n", id);
	if (rs == MISANSWER)  
	    pop3_quit(gm);
	return rs;
    }
    
    sprintf(buf, "PASS %s", pass);
    if ((rs = pop3_transact(gm->socket, buf, SVRBUF)) != OK)  {
	dbgPrintf("User %s certificate failed.\n", id);
	if ( rs == MISANSWER )
	    pop3_quit(gm);
    }
    
    return rs;
}
 
static int pop3_check(GMAIL *gm)
{
    char  buf[SVRBUF], *idx[5];
    int   rs, i;
    
    
    /* use STAT command to check the mailbox. Get the mails
       number and the total size of them */
    
    strcpy(buf, "STAT");
    rs = pop3_transact(gm->socket, buf, SVRBUF);
    if (rs != OK)  
	return rs;
    
    ziptoken(buf, idx, 3, " ");
    gm->msgs = atoi(idx[1]);
    gm->volm = atoi(idx[2]);
    if (gm->msgs == 0)  
	return OK;
    
    /* now check the possibility of TOP command */
    
    strcpy(buf, "TOP 1 0");
    rs = pop3_transact(gm->socket, buf, SVRBUF);
    if (rs != OK)  {
	logPrintf(LOG_WARN, "TOP command isn't supported by peer.\n");
	gm->distop = 1;
    }
    
    while(1)  {
	rs = sock_readline(gm->socket, buf, SVRBUF, timeout);
	if ((rs != OK) || !strcmp(buf, ".\n"))  
	    break;
    }		
    
    /* if there were mails existed,  use LIST command to get
       each letter's size */
    
    gm->mlist = malloc(gm->msgs * sizeof(int));
    if (gm->mlist == NULL)  {
	dbgPrintf("pop3_check: out of memory.\n");
	return FAILURE;
    }
    
    strcpy(buf, "LIST");
    rs = pop3_transact(gm->socket, buf, SVRBUF);
    if (rs != OK)  
	return rs;

    for (i = 0; i < gm->msgs; i++)  {
	    
    	rs = sock_readall(gm->socket, buf, SVRBUF, timeout);
    	if (rs != OK)  
	    return rs;
	
    	if (!strcmp(buf, ".\n"))  
	    break;
    	if (verbose)  
	    conPrintf("<< %s", buf);
	
	ziptoken(buf, idx, 2, " ");
    	gm->mlist[i] = atoi(idx[1]);
    }
    return sock_readall(gm->socket, buf, SVRBUF, timeout);
}

/************************** RETR ****************************/

static int pop3_retr_one(GMAIL *gm, int id)
{
    char   buf[SVRBUF];
    int    rs;
    
    sprintf(buf, "RETR %d", id);
    rs = pop3_transact(gm->socket, buf, SVRBUF);
    if (rs != OK)  {
	dbgPrintf("gmail: fetching mail failed. (%d/%d)\n", id, gm->msgs);
    	return rs;
    }
    	
    fprintf(gm->fp, "From - %s\n", timenow());
    prog_init(gm->mlist[id - 1], 40);
    while(1)  {
	rs = sock_readline(gm->socket, buf, SVRBUF, timeout);
    	if (rs != OK)  {
	    dbgPrintf("gmail: transfer broken. (%d/%d)\n", id, gm->msgs);
    	    return rs;
    	}
    	if (!strcmp(buf, ".\n"))  
	    break;
    		
    	prog_do(strlen(buf));
    	if (*buf == '.')  
	    fputs(buf + 1, gm->fp);
    	else  
	    fputs(buf, gm->fp);
    }
    fputs("\n", gm->fp);
    prog_end();
    	
    return OK;
}


static int pop3_process_spam(GMAIL *gm, Spam *sc, Head *head, int sn)
{
    fd_set rfds;
    struct timeval tv;
    char   buf[32];
    int    i, rs;

    conPrintf("\n");
    if (head->num == 0) {
	pop3_list_head(gm, head, sn, TOP_SMART);
    } else {
	conPrintf("Message %d# of %d, %d bytes\n",
		sn + 1, gm->msgs, gm->mlist[sn]);
	pop3_head_digest(gm, head);
    }
    
    if (sc->quota)
	conPrintf("Warning: Message oversized. (%d > %d)\n", 
			gm->mlist[sn], sc->quota);
    if (sc->from_reg)
	conPrintf("Warning: Unwelcome address: %s\n", sc->from);
    if (sc->sub_reg)
	conPrintf("Warning: Unwelcome subject: %s\n", sc->subject);
    
    switch (sc->action)  {
    case SPAM_KEEP:
	conPrintf("Leave this message untouched.\n");
	break;

    case SPAM_DELE:
	conPrintf("Remove this message from server.\n");
	pop3_dele_one(gm, sn + 1);
	break;

    case SPAM_ASK:
    default:
	conPrintf("Message processing? (Keep, Remove, Download) ");
	tcflush(0, TCIFLUSH);
	FD_ZERO(&rfds);
	FD_SET(0, &rfds);
	tv.tv_sec = 30;
	tv.tv_usec = 0;
	rs = select(1, &rfds, NULL, NULL, &tv);
	if (rs) {
	    rs = read(0, buf, 32);
	    for (i = 0; (i < rs) && isspace(buf[i]); i++);
	    if ((buf[i] == 'R') || (buf[i] == 'r'))
		pop3_dele_one(gm, sn + 1);
	    else if ((buf[i] == 'D') || (buf[i] == 'd'))
		return SPAM_FETCH;
	    
	    tcflush(0, TCIFLUSH);
	} else 
	    conPrintf("Keep\n");
	break;
    }

    return OK;
}
    

/* 0: fine; Others: reject it */

static int pop3_spam_match(GMAIL *gm, Head *head, int sn)
{
    char  buf[SVRBUF];
    int   rs;
    Spam  *sc;
    
    if (!gm->slnk)  
	return OK;		/* no spam message check required */

    /* first check the message size quota. I desolate it from commom match
     * work because it does not need to download the message head. */

    sc = spam_match_size(gm->slnk, gm->mlist[sn]);
    if (sc != NULL)  {
	head_clean(head);
	if (pop3_process_spam(gm, sc, head, sn) == SPAM_FETCH)
	    return OK;
	return SPAM_MSG;
    }
    
    if (!gm->topspam)
	return OK;
    
    /* here goes the routine work: download the message head, search for any 
     * unwelcome informations. */
    
    if (gm->distop)
	return OK;	/* server does not support TOP command */
    
    sprintf(buf, "TOP %d 0", sn + 1);
    rs = pop3_transact(gm->socket, buf, SVRBUF);
    if (rs != OK)  
	return FAILURE;

    head_clean(head);
    while(1)  {
	rs = sock_readline(gm->socket, buf, SVRBUF, timeout);
	if (rs != OK)  {
	    dbgPrintf("pop3_spam_match: transfer is broken by peer.\n");
	    return FAILURE;
	}
	if (!strcmp(buf, ".\n"))  break;
	
	head_append(head, buf);
    }

    sc = spam_match_chain(gm->slnk, head, gm->mlist[sn]);
    if (sc != NULL)  {
	if (pop3_process_spam(gm, sc, head, sn) == SPAM_FETCH)
	    return OK;
	return SPAM_MSG;
    }
    
    return OK;
}


static int pop3_retr(GMAIL *gm, Head *head)
{
    int  i, rs;
   
    if ((gm->fp = fopen(gm->mbox, "a+")) == NULL)  {
	dbgPrintf("cannot open folder [%s]\n", gm->mbox);
	return OPENFILERR;
    }
    
    for (i = 0; i < gm->msgs; i++)  {
	    
    	if ( !isinset(i + 1, gm->attr->cp_set) )  
	    continue;
    	
    	conPrintf("fetching %3d/%d (%9d) ", i + 1, gm->msgs, gm->mlist[i]);
    	if (verbose)  
	    conPrintf("\n");
    	
	rs = pop3_spam_match(gm, head, i);
	if (rs != OK)  
	    continue;		/* message rejected */
    	
	if (pop3_retr_one(gm, i + 1) != OK)  
	    break;
	
	if (gm->act & GMAIL_DELE)  
	    pop3_dele_one(gm, i + 1);
    }
    pfclose(gm->fp);
    return OK;
}

/******************* LIST *************************/

static int pop3_list_basic(GMAIL *gm)
{
    int  i;
    
    for (i = 0; i < gm->msgs; i++)  {
	conPrintf("%3d# of %d, %9d bytes\n", i + 1, gm->msgs, gm->mlist[i]);
    }
    return OK;
}

/* display message head in a single line, include series number, from address
 * and subject */

static void pop3_head_info(GMAIL *gm, Head *head, int sn)
{	
    HLin   *hl;
    char   buf[SVRBUF];
    int	   pro[4];
    
    /* calculate each part's proportion */

    pro[0] = gm->attr->cols / 20;
    pro[1] = gm->attr->cols / 10 + 1;
    pro[2] = gm->attr->cols / 3 - 1;
    pro[3] = gm->attr->cols - pro[0] - pro[1] - pro[2] - 4;
    
    /* check the from address of the message */
    
    hl = head_check(head, "from");
    if (hl == NULL)  
	hl = head_check(head, "sender");
    if (hl == NULL)  
	hl = head_check(head, "reply");

    if (hl == NULL)  
	siprintf(buf, SVRBUF, "%d#%d %s", 
		pro[0], sn + 1, pro[1], gm->mlist[sn], -pro[2], "Anomymous");
    else {
	compress_space(hl->body);
	siprintf(buf, SVRBUF, "%d#%d %s", 
		pro[0], sn + 1, pro[1], gm->mlist[sn], -pro[2], hl->body);
    }
    conPrintf(buf);
    
    if ((hl = head_check(head, "subject")) != NULL)  {
	compress_space(hl->body);
	siprintf(buf, SVRBUF, " %s\n", -pro[3], hl->body);
	conPrintf(buf);
    } else 
	conPrintf("\n");
}


static void pop3_head_digest(GMAIL *gm, Head *head)
{
    HLin  *hl;
    int   i, rs;

    for (i = rs = 0; gm->attr->hdmsk[i]; i++)  {
	if ((hl = head_check(head, gm->attr->hdmsk[i])) != NULL)  {
	   fputs(hl->headin, stdout);
	   rs++;
	}
    }
    if (rs == 0)
	fputs("Warning: mystery message source!\n", stdout);

    fflush(stdout);
}

/* ctl is same as topstyle. If the ctl is less than 0, it uses predefined 
 * topstyle instead. In fact it's a leak way to overide the predefined one */

static int pop3_list_head(GMAIL *gm, Head *head, int sn, int ctl)
{
    char   buf[SVRBUF];
    int	   rs;
    
    /* it assumes that the TOP command is supported by the sevice provider */
    
    sprintf(buf, "TOP %d 0", sn + 1);
    rs = pop3_transact(gm->socket, buf, SVRBUF);
    if (rs != OK)  
	return rs;
    
    /* download the message head and store them in the Head object */
    
    head_clean(head);
    while(1)  {
        rs = sock_readline(gm->socket, buf, SVRBUF, timeout);
        if (rs != OK)  {
	    dbgPrintf("pop3_list_head: transfer is broken by peer.\n");
	    return rs;
	}
    	if (!strcmp(buf, ".\n"))  
	    break;
    	
	head_append(head, buf);
    }

    if (ctl < 0)
	ctl = gm->attr->tops;
    
    switch (ctl) {
    case TOP_ALL:
	conPrintf("\nMessage %d# of %d, %d bytes\n", 
    			sn + 1, gm->msgs, gm->mlist[sn]);
	head_dump(head, stdout, 1);
	conPrintf("\n");
	break;
	
    case TOP_SMART:
	conPrintf("\nMessage %d# of %d, %d bytes\n", 
    			sn + 1, gm->msgs, gm->mlist[sn]);
	pop3_head_digest(gm, head);
	break;

    case TOP_BRIEF:
	pop3_head_info(gm, head, sn);
	break;
    }
    
    return OK;
}


static int pop3_list_top(GMAIL *gm, Head *head)
{
    int	   rs, i;
    
    if (gm->distop) 
	return pop3_list_basic(gm);
	   
    for (i = rs = 0; i < gm->msgs; i++)  {
	    
    	if (!isinset(i + 1, gm->attr->cp_set))  
	    continue;

	rs = pop3_list_head(gm, head, i, -1);
	if (rs != OK) 
	    break;
    }
    return rs;
}


/************************* delete ****************************/

static int pop3_dele_one(GMAIL *gm, int id)
{
    char buf[SVRBUF];
    int  rs;
    
    sprintf(buf, "DELE %d", id);
    rs = pop3_transact(gm->socket, buf, SVRBUF);
    if (rs != OK) 
	dbgPrintf("pop3_dele_one: delete message %d failed.\n", id);
    
    return rs;
}

static int pop3_dele(GMAIL *gm)
{
    int  rs, cnt, i;
    
    /* checking procession set, it should be existed */
    /* if (!gm->attr->cp_set)  return OK; */
    
    for (i = cnt = 0; i < gm->msgs; i++)  {
	    
    	if (!isinset(i + 1, gm->attr->cp_set))  
	    continue;
	
    	conPrintf("remove %3d of %d, %d bytes\n", 
			i + 1, gm->msgs, gm->mlist[i]);
    
	rs = pop3_dele_one(gm, i + 1);	
    	if (rs != OK)  
	    break; 
  
    	cnt++;
    }
    conPrintf("Remove %d messages from [%s]\n", cnt, gm->host);
    return OK;
}

static int pop3_quit(GMAIL *gm)
{
    char  buf[SVRBUF];
    int   rs;

    if (verbose)  
	conPrintf("-> QUIT\n");
    
    sprintf(buf, "QUIT\r\n");
    
    rs = sock_write(gm->socket, buf, strlen(buf), timeout);
    if (rs != OK)  {
	dbgPrintf("pop3_quit: write socket failed [%s]\n", "QUIT");
	return rs;
    }
   /* 
    rs = sock_readall(gm->socket, buf, SVRBUF, timeout);
    if (rs != OK)  {
	dbgPrintf("pop3_quit: read socket failed.\n");
	return rs;
    }
    if (verbose || (*buf != '+'))  conPrintf("<< %s", buf);
    */
    
    return OK;
}
    
static int pop3_transact(int socket, char *cmd, int len)
{
    char  buf[SVRBUF];
    int   rs;
    
    if (verbose)  
	conPrintf("-> %s\n", cmd);
    
    sprintf(buf, "%s\r\n", cmd);
    rs = sock_write(socket, buf, strlen(buf), timeout);
    if (rs != OK)  {
	dbgPrintf("pop3_transact: Write socket failed [%s]\n", cmd);
	return rs;
    }
    
    rs = sock_readall(socket, buf, SVRBUF, timeout);
    if (rs != OK)  {
	dbgPrintf("pop3_transact: Read socket failed.\n");
	return rs;
    }
    if (verbose)  
	conPrintf("<< %s", buf);
    
    if (*buf != '+')  {
    	logPrintf(LOG_WARN, "pop3_transact: %s failed: %s", cmd, buf);
    	strcpy(buf, "QUIT\r\n");
    	rs = sock_write(socket, buf, strlen(buf), timeout);
    	if (rs != OK)  
	    return rs;
    	rs = sock_readall(socket, buf, SVRBUF, timeout);
    	if (rs != OK)  
	    return rs;
    	return MISANSWER;
    }
    strncpy(cmd, buf, len);
    return OK;
}	



